'''
Welcome to PalmettoBUG! 

PalmettoBUG's analysis functions are available either thorugh the GUI, or through scripting.
To launch the GUI, run the command 'palmettobug' in a CLI environment where PalmettoBUG is installed, or use the function 
PalmettoBUG.run_GUI() in a script.


License: GPL3
Author: Ben Caiello
Institution: FlowCytometry and Cell Sorting Shared Resource of the Hollings Cancer Center at the Medical University of South Carolina


This script contains the code for exporting the various functions / classes of PalmettoBUG's non-GUI API
'''

from .Executable import run_GUI    ## might as well create an additional launch point for the GUI program

from .Entrypoint.app_and_entry import fetch_CyTOF_example, fetch_IMC_example
from .Entrypoint.bead_norm import CyTOF_bead_normalize

from .ImageProcessing.ImageAnalysisClass import (ImageAnalysis, 
                                               mask_expand, 
                                               imc_entrypoint, 
                                               read_txt_file, 
                                               txt_folder_to_tiff_folder,
                                               setup_for_FCS)

from .Pixel_Classification.Classifiers import SupervisedClassifier, UnsupervisedClassifier, plot_pixel_heatmap, plot_class_centers, segment_class_map_folder
from .Pixel_Classification.use_classifiers import (plot_classes,
                                                   merge_classes, 
                                                   merge_folder,
                                                   slice_folder,
                                                   mode_classify_folder,
                                                   secondary_flowsom,
                                                   classify_from_secondary_flowsom,
                                                   extend_masks_folder, 
                                                   )
                 
from .Analysis_functions.WholeClassAnalysis import WholeClassAnalysis
from .Analysis_functions.Analysis import Analysis
#from .Analysis_functions.SpatialANOVA import SpatialANOVA, plot_spatial_stat_heatmap, do_functional_ANOVA, do_K_L_g
from .Analysis_functions.SpatialAnalysis import SpatialAnalysis

from .Utils.sharedClasses import TableLaunch_nonGUI as TableLaunch
from .Utils.sharedClasses import run_napari


__all__ = ["run_GUI",
           "CyTOF_bead_normalize",
           "ImageAnalysis",
           "mask_expand",
           "imc_entrypoint",
           "read_txt_file",
           "txt_folder_to_tiff_folder",
           "setup_for_FCS",
           "SupervisedClassifier",
           "UnsupervisedClassifier",
           "plot_pixel_heatmap",
           "plot_class_centers",
           "plot_classes",
           "segment_class_map_folder",
           "extend_masks_folder",
           "slice_folder",
           "secondary_flowsom",
           "classify_from_secondary_flowsom",
           "mode_classify_folder",
           "merge_classes",
           "merge_folder",
           "WholeClassAnalysis",
           "Analysis",
           "SpatialAnalysis",
           "TableLaunch",
           "run_napari",
           "fetch_CyTOF_example",
           "fetch_IMC_example",
           "print_license",
           "print_3rd_party_license_info"
           ]

homedir = __file__.replace("\\","/")
homedir = homedir[:(homedir.rfind("/"))]

print("The PalmettoBUG package is copyrighted 2024-2025 by the Medical University of South Carolina and licensed under the GPL-3 license."
    "\nIt is free & open source software, can  be redistributed in compliance with the GPL3 license, and comes with absolutely no warranty."
    "\nIn python, use palmettobug.print_license() to see the license, or use palmettobug.print_3rd_party_license_info() to print information"
    "\nabout the licenses and copyright of 3rd party software used in PalmettoBUG itself or in the creation of PalmettoBUG.")

def print_license():
    license_dir = homedir + "/Assets/LICENSE.txt"
    with open(license_dir) as file:
        license = file.read()
    print(license)


def print_3rd_party_license_info():
    license_dir = homedir + "/Assets/Other_License_Details.txt"
    with open(license_dir) as file:
        license = file.read()
    print(license)



# ruff: noqa
'''
This file is an "vendorizing" of the sigfig package: https://github.com/drakegroup/sigfig

(
MIT License

Copyright (c) 2022 drakegroup

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
)

Changes:
    -- removed a mysterious symbol from the first comment in the copied code (maybe a ctrl+C / ctrl+V artifact? Not visible on github where it was copied from)
    -- replaced assert --> if not...raise 
    -- add __all__ for docs

Note on linting: Like with most of the libraries I semi-vendorized, I have the `ruff: noqa` comment at the top to block linting. Howveer, also like the other vendorizing files,
I did see the results of an initial linting by ruff (3-28-25) --> however, unlike the other vendorized files, there were questionable patterns that might deserve editing at some point:
    --> using `==` instead of `is` for comparisons to True / False
    --> using `==` instead of isinstance() for types
    --> using bare exceptions
'''
__all__ = []
######### Begin copied code:

#!/usr/bin/python
# -*- coding: utf-8 -*-

from decimal import Decimal
from copy import deepcopy
from sortedcontainers import SortedDict
from warnings import warn, filterwarnings, resetwarnings

import numbers

from inspect import currentframe, getfile
def _warn_stacklevel(best_guess=2):
    '''Helper function to associate warnings with caller outside this module.

    Args:
        best_guess: stack depth estimate in case the Python interpreter doesn't support stack introspection.
    Returns:
        integer stack level for warnings.warn() function.

    Note:
    - Python 3.12 should simplify this process with 'skip_file_prefixes' once the path normalization bugs are fixed.
    '''
    stacklevel = 0
    try:
        frame = currentframe()
        while True:
            if frame is None or getfile(frame) != __file__:
                return stacklevel
            frame = frame.f_back
            stacklevel += 1
    except:
        return best_guess

_manual_settings = {}
_default_settings = {
    'spacing': 0.1,
    'spacer': '',
    'decimal': '.',
    'separator': ' ± ',
    'cutoff': 9,
    'prefix': False,
    'exponent': 'E',
    'round_by_sigfigs': False,
    'round_by_decimals': True,
    'given_sigfigs': 0}
_major_prefixes = {
    24: ['Y', 'septillion'],
    21: ['Z', 'sextillion'],
    18: ['E', 'quintillion'],
    15: ['P', 'quadrillion'],
    12: ['T', 'trillion'],
    9: ['G', 'billion'],
    6: ['M', 'million'],
    3: ['k', 'thousand'],
    0: ['',''],
    -3: ['m', 'thousandth'],
    -6: ['μ', 'millionth'],
    -9: ['n', 'billionth'],
    -12: ['p', 'trillionth'],
    -15: ['f', 'quadrillionth'],
    -18: ['a', 'quintillionth'],
    -21: ['z', 'sextillionth'],
    -24: ['y', 'septillionth']}
_minor_prefixes = {
    2: ['h', 'hundred'],
    1: ['da', 'ten'],
    -1: ['d', 'tenth'],
    -2: ['c', 'hundredth']}

class _Number:
    '''
    Private data structure for storing & manipulating numbers

    Attributes:
        .map: - dict for number storage
              - maps 10's power (key) to numeric value (value)
              - i.e. 3.14 => .map = {0:3, -1: 1, -2: 4}
    Attributes/Mehtods for inspection (getting values):
        .sign:     string of either '+' or '-', denoting sign of stored number
        .nan:      bool True/False depending on whether number is NaN
        .positive: bool True/False depending on number's sign
        .negative: bool True/False depending on number's sign
        .has_uncertainty: bool True/False depending on whether there is an associated uncertainty with this number
        .max_power(): returns integer corresponding to number's highest populated 10's power
        .min_power(): returns integer corresponding to number's lowest populated 10's power
        .decimate(dict, _Number, bool, bool): returns string of all digits in given format
            specifying spacing and non-standard decimal point, optional sign
            optionally embedded uncertainty, and optional leading/trailing zeros
        .output(type): returns number in given type
    Methods for manipulation (changing the value):
        .set_sign(str): function used to change/set the number's sign by passing '-' or '+'
                        so that .positive, .negative, .sign don't need manual updating
        .increment_power_by(int): (de)increments all keys in .map by given value
        .round_by_decimals(int):  performs rounding operation to the given 10's power
        .prefixify(str):          converts to Scientific or Engineering notation with optional SI prefix
    '''
    def __init__(self):
        self.set_sign('+')
        self.prefix = ''
        self.has_uncertainty = False
        self.map = SortedDict()
        self.zero = False
        self.nan = False
    def set_sign(self, sign='+'):
        '''sets the number's sign'''
        if sign == '+':
            self.negative = False
            self.positive = True
            self.sign = '+'
        elif sign == '-':
            self.negative = True
            self.positive = False
            self.sign = '-'
        else:
            warn('sign must be "+" or "-", assuming positive', stacklevel=_warn_stacklevel(4))
            self.negative = False
            self.positive = True
            self.set_sign('+')
    def max_power(self):
        '''returns integer corresponding to number's highest populated 10's power'''
        return max(self.map)
    def min_power(self):
        '''returns integer corresponding to number's lowest populated 10's power'''
        return min(self.map)
    def increment_power_by(self, n):
        '''(de)increments all keys in .map'''
        tmp = SortedDict()
        for key in self.map:
            tmp[key + n] = self.map[key]
        self.map = tmp
    def round_by_decimals(self, decimals):
        '''performs rounding operation to the given 10's power'''
        last_power = -decimals
        tmp_map = SortedDict({self.max_power(): 0})
        tmp_map = SortedDict({last_power: 0})
        last_tmp_power = self.max_power()
        for key in range(self.max_power(), min(-decimals - 1, -1), -1):
            if key < last_power and key >= 0 or key not in self.map:
                float(0)
            #if key < last_power and key < 0:
            #    Decimal(3)
            #    continue
            else:
                last_tmp_power = key
                tmp_map[key] = self.map[key]
        if last_power == self.max_power() + 1:
            if self.map[self.max_power()] >= 5:
                tmp_map[last_power] = 1
            else:
                self.zero = True
                tmp_map = SortedDict({last_power:0})
                self.set_sign('+')
        elif last_power > self.max_power():
            self.zero = True
            tmp_map = SortedDict({last_power:0})
            self.set_sign('+')
        elif last_tmp_power - 1 in self.map and self.map[last_tmp_power - 1] >= 5:
            tmp_map[last_tmp_power] += 1
            tmp_power = last_tmp_power
            while tmp_map[tmp_power] == 10:
                tmp_map[tmp_power] = 0
                tmp_power += 1
                if tmp_power not in tmp_map:
                    tmp_map[tmp_power] = 1
                else:
                    tmp_map[tmp_power] += 1
        self.map = tmp_map
        '''if self.map == SortedDict({0:0}):
            self.zero = True
            self.set_sign('+')
            if self.has_uncertainty:
                if last_power < 0:
                    for p in range(-1, last_power - 1, -1):
                        self.map[p] = 0
                elif last_power > 0:
                    self.map = SortedDict({last_power:0})'''
    def decimate(self, format, unc=None, zeropadding=True, sign=True, units=''):
        '''
        returns string of all digits in given format {spacing, spacer, decimal},
        with unc=_Number for embedded uncertainty, and optional leading/trailing zeros & sign
        '''
        top = self.max_power()
        bot = self.min_power()
        if zeropadding:
            top = max(top, 0)
            bot = min(bot, 0)
        if self.zero and not unc and top > 0:
            top = 0
        output = []
        if sign and self.negative:
            output.append('-')
        for p in range(top, bot - 1, -1):
            try:
                output.append(str(self.map[p]))
            except:
                output.append('0')
            if p == self.min_power() and unc:
                output.append('('+unc.decimate(format, zeropadding=False, sign=False)+')')
            if p != bot:
                if p == 0:
                    output.append(format['decimal'])
                elif p % format['spacing'] == 0:
                    output.append(format['spacer'])
        return ''.join(output) + units
    @staticmethod
    def _int(num):
        return int(float(num))
    def output(self, output_type):
        '''returns number in given type'''
        no_formatting = {'decimal': '', 'spacer': '', 'spacing': 0.1}
        num = self.decimate(no_formatting, zeropadding=False) or "0"
        output_type = self._int if output_type == int else output_type
        return output_type(f"{num}E{self.min_power()}")
    def __gt__(self, other):
        if self.max_power() > other.max_power():
            return True
        for p in range(self.max_power(), self.min_power() - 1, -1):
            if p not in other.map:
                if not self.map[p]:
                    continue
                return True
            if self.map[p] > other.map[p]:
                return True
            if self.map[p] < other.map[p]:
                return False
        if other.min_power() > self.min_power():
            return False
        return False
    def prefixify(self, prefix, exponent):
        '''converts to Engineering/Scientific notation with optional SI prefix'''
        #self.prefix = 'XXX'
        #return None
        self.prefix = ''
        power_shift = 0
        if prefix == 'sci':
            p = self.max_power()
            self.prefix = exponent + str(p)
            self.increment_power_by(-p)
            power_shift += -p
            return power_shift
        elif prefix == 'eng':
            p0 = 0
            if self.max_power() < 0:
                p0 = 2
            p = int((self.max_power() - p0)/3)*3
            self.prefix = exponent + str(p)
            self.increment_power_by(-p)
            power_shift += -p
            return power_shift
        elif prefix in ['all', 'minor']:
            prefixes = SortedDict(_major_prefixes+_minor_prefixes)
            p = self.min_power()
            while p >= min(prefixes):
                if p in prefixes:
                    break
                p -= 1
        else:
            #prefix in [True, 'major', 'eng']:
            prefixes = _major_prefixes
            #p = int((self.min_power()-1)/3)*3
            p = self.max_power()
            while p > max(prefixes)+3:
                self.prefix += prefixes[max(prefixes)][0]
                self.increment_power_by(-max(prefixes))
                power_shift += -max(prefixes)
                p -= max(prefixes)
            while p < min(prefixes):
                self.prefix += prefixes[min(prefixes)][0]
                self.increment_power_by(-min(prefixes))
                power_shift += -min(prefixes)
                p -= min(prefixes)
            p0 = 0
            if self.max_power() < 0:
                p0 = 2
            p = int((self.max_power() - p0)/3)*3
        self.prefix = prefixes[p][0] + self.prefix
        self.increment_power_by(-p)
        power_shift += -p
        return power_shift

def _arguments_parse(args, kwargs):
    '''Private function for use only in round() function:
    Deciphers user intent based on given inputs along with preset defaults
    which returns actionable and summarized useful variables in a dict.
    '''
    given = {'reset_warnings': False}

    if any([w in kwargs for w in ('warn', 'warning', 'warnings')]):
        warning = kwargs.get('warn') or kwargs.get('warning') or kwargs.get('warnings')
        if not warning:
            filterwarnings('ignore')
        elif warning == True:
            resetwarnings()
        else:
            resetwarnings()
            warn(f'warnings argument expected to be True, False, or "once". Got "{warning}"', stacklevel=_warn_stacklevel(3))
        given['reset_warnings'] = True

    types = (numbers.Number, str, Decimal, _Number, type(None))
    given['output_type'] = type(args[0])
    if not isinstance(args[0], types):
        raise TypeError(f'Invalid input type of {type(args[0])}, expecting 1 of {types}')
    given['num'] = _num_parse(args[0])
    if len(args) >= 2:
        if type(args[1]) == int:
            given['sigfigs'] = args[1]
            if given['sigfigs'] < 1:
                warn('cannot have less that 1 significant figure, setting to 1', stacklevel=_warn_stacklevel(3))
                given['sigfigs'] = 1
        elif args[1] != args[1]:
            warn(f'Ignoring 2nd argument "{args[1]}". invalid uncertainty, expecting number', stacklevel=_warn_stacklevel(3))
        else:
            given['uncertainty'] = _num_parse(args[1])
            given['output_type'] = str
    if len(args) > 2:
        warn(f"last {int(len(args) - 3)} argument(s) discarded/ignored", stacklevel=_warn_stacklevel(3))
    
    for key in _manual_settings:
        given[key] = _manual_settings[key]

    keys = {'separator', 'separation', 'sep', 'format', 'sigfigs', 's', 'decimals', 'd', 'uncertainty', 'u', 'cutoff', 'spacing', 'spacer', 'decimal', 'output_type', 'output', 'type', 'style', 'prefix', 'exponent', 'notation', 'form', 'crop'}
    for key in kwargs:
        val = kwargs[key]
        if key not in keys:
            warn(f'unrecognized argument, skipping {key}={val}', stacklevel=_warn_stacklevel(3))
            continue
        if key in given and key not in _manual_settings:
            None
            #warn("overwriting %s=%s with %s=%s" % (key, given[key], key, val))
        if key in {'sigfigs', 's', 'decimals', 'd', 'cutoff', 'crop'}:
            shortcuts = {'s':'sigfigs', 'd':'decimals'}
            key = shortcuts[key] if key in shortcuts else key
            try:
                if key == 'crop':
                    key = 'cutoff'
                if not isinstance(val, numbers.Integral):
                    warn(f'use integer type for {key} argument', stacklevel=_warn_stacklevel(3))
                if key in {'cutoff', 'crop'} and int(val) < 9:
                    warn('cutoff/crop cannot be < 9, setting to 9', stacklevel=_warn_stacklevel(3))
                    val = 9
                if key == 'sigfigs' and int(val) < 1:
                    warn('cannot have less that 1 significant figure, setting to 1', stacklevel=_warn_stacklevel(3))
                    val = 1
                given[key] = int(val)
            except:
                warn(f'Ignoring {key}={val}, invalid type of {type(val)}, expecting integer type', stacklevel=_warn_stacklevel(3))
        elif key in {'uncertainty', 'u'}:
            try:
                if not(val == val):
                    raise Exception
                given['uncertainty'] = _num_parse(val)
                given['output_type'] = str
            except:
                warn(f"Ignoring {key}={val}. invalid uncertainty, expecting number", stacklevel=_warn_stacklevel(3))
        elif key == 'prefix':
            if type(val) == bool or val in {'major', 'sci', 'eng'}:
                given[key] = val
            elif val in ['minor', 'all']:
                given['prefix'] = 'all'
            else:
                prefixes = { True, False, 'major', 'minor', 'sci', 'eng', 'all'}
                warn(f"Ignoring {key}={val}, invalid prefix setting, expecting 1 of: {prefixes}", stacklevel=_warn_stacklevel(3))
                continue
            given['output_type'] = str
        elif key in {'spacer', 'decimal'}:
            given[key] = str(val)
            given['output_type'] = str
        elif key == 'spacing':
            given['spacing'] = int(val)
            given['output_type'] = str
        elif key in {'sep', 'separation', 'separator'}:
            if val == 'external_brackets':
                given['separator'] = 'brackets'
                given['external_brackets'] = True
                continue
            elif val == 'brackets':
                given['separator'] = 'brackets'
            elif val in {tuple, list}:
                given['output'] = val
                given['output_type'] = type(args[0])
            elif val == 'tuple':
                given['output'] = tuple
                given['output_type'] = type(args[0])
            elif val == 'list':
                given['output'] = list
                given['output_type'] = type(args[0])
            given['separator'] = str(val)
        elif key in ['format', 'style', 'output', 'type', 'output_type', 'notation', 'form']:
            given['output_type'] = str
            #warning might be warranted if output_type previously specified
            properties = ['spacing', 'spacer', 'decimal', 'separator', 'cutoff', 'prefix', 'form']
            formats = {'English': [3, ',', '.', ' ± ', 9, False,  '#,###,###.## ± 0.#'],
                       'French':  [3, ' ', ',', ' ± ', 99, False, '# ### ###,## ± 0,##'],
                       'other':   [3, '.', ',', ' ± ', 99, False, '#.###.###,## ± 0,##'],
                       'PDG':     [.1, '', '.', ' ± ', 35, False,  '# ### ###.##(##)'],
                       'Drake':   [3, ' ', '.', 'brackets', 29, False, '# ### ###.##(##)'],
                       'sci':     [.1, '', '.', ' ± ', 9, 'sci', '# ### ###.##(##)'],
                       'eng':     [.1, '', '.', ' ± ', 9, 'eng', '# ### ###.##(##)'],
                       'std':     [.1, '', '.', ' ± ', 9, False, '# ### ###.##(##)']}
            outputs = {'+-', 'map'}
            notations = {'sci', 'scientific', 'eng', 'engineering', 'std', 'standard'}
            if val in notations:
                val = val[:3]
                if val == 'sta':
                    val = 'std'
                for i, prop in enumerate(properties):
                    #if prop in given and key not in _manual_settings:
                    if prop in given:
                        continue
                        #warn("overwriting %s=%s with %s=%s" % (prop, given[prop], prop, formats[val][i]))
                    given[prop] = formats[val][i]
            elif val in formats:
                for i, prop in enumerate(properties):
                    if prop in given and key not in _manual_settings:
                        None
                        #warn("overwriting %s=%s with %s=%s" % (prop, given[prop], prop, formats[val][i]))
                    given[prop] = formats[val][i]
            elif isinstance(val, type) and issubclass(val, types):
                given['output_type'] = val
                if 'prefix' in given:
                    del given['prefix']
            elif val in outputs:
                if val == '+-':
                    given['separator'] = ' ± '
                else:
                    given['output'] = val
            else:
                warn(f"expected format of {[f for f in formats] + [o for o in outputs]}, ignoring format of {val}", stacklevel=_warn_stacklevel(3))
                given['output_type'] = type(args[0])
        else:
            given[key] = val

    if 'sigfigs' in given and 'decimals' in given:
        warn('Cannot round by both sigfigs & decimals, ignoring decimal constraint', stacklevel=_warn_stacklevel(3))
        del given['decimals']
    if 'uncertainty' in given and any([x in given for x in ['sigfigs', 'decimals', 'arg2']]):
        warn(
            'Cannot round by both uncertainty & decimals/sigfigs simultaneously, ignoring decimals &/or sigfigs.  Use seperate calls to round() function for seperate roundings.',
            stacklevel=_warn_stacklevel(3)
        )
        for prop in ['sigfigs', 'decimals', 'arg2']:
            if prop in given:
                del given[prop]
    if 'arg2' in given:
        if 'sigfigs' in given or 'decimals' in given:
            warn('Invalid 2nd argument, ignoring. "sigfigs" or "decimals" given in keyword arguments', stacklevel=_warn_stacklevel(3))
        elif _default_settings['round_by_sigfigs']:
            given['sigfigs'] = given['arg2']
        else:
            given['decimals'] = given['arg2']
        del given['arg2']
    if not issubclass(given['output_type'], (numbers.Real, Decimal, _Number)):
        given['format'] = {}
        if 'spacer' in given and 'spacing' not in given:
            given['spacing'] = 3
        if 'spacing' in given and 'spacer' not in given:
            given['spacer'] = ","
        for prop in {'decimal', 'spacer', 'spacing'}:
            if prop in given:
                val = given[prop]
                del given[prop]
            elif prop in _manual_settings:
                val = _manual_settings[prop]
            else:
                val = _default_settings[prop]
            given['format'][prop] = val
    for prop in ['separator', 'prefix', 'exponent']:
        if prop in given:
            continue
        elif prop in _manual_settings:
            given[prop] = _manual_settings[prop]
        else:
            given[prop] = _default_settings[prop]

    return given
def _num_parse(num):
    '''Private function for use only in round()'s _arguments_parse() function:
    Translates given number of any type into returned _Number data structure

    Parsing Algorythm [O(N)]:
    - convert to string
    - characters are analyzed sequentially in a KMP-like state graph. ie:
       number: -325.7854E-5 
       state:  ABBBBCCCCDEE
    - ValueError is raised if input number cannot be deciphered
    '''
    global number, i, n, negative_exp, exp
    number = _Number()
    i = 0
    n = 0
    negative_exp = False
    exp = 0

    if type(num) == type(number):
        return deepcopy(num)
    if num is None:
        warn('no number provided, assuming zero (0)', stacklevel=_warn_stacklevel(4))
        number.map[0] = 0
        number.zero = True
        return number
    if num != num:
        warn('given input is not a number (NaN)')
        number.nan = True
        number.map['NaN'] = num
        return number
    num = str(num)

    digits = set([str(a) for a in range(10)])
    exponents = set(['E', 'e', 'D', 'd', 'Q', 'q'])

    def A(num):
        global number, i, n
        i += 1
        if not num or num in '.-+':
            warn('no number provided, assuming zero (0)', stacklevel=_warn_stacklevel(4))
            number.map[0] = 0
            number.zero = True
            return None
        elif num[0] in exponents:
            warn('no number provided, assuming zero (0)', stacklevel=_warn_stacklevel(4))
            number.map[0] = 0
            number.zero = True
            D(num[1:])
        elif num[0] in '+-':
            number.set_sign(num[0])
            B(num[1:])
        elif num[0] in '.':
            C(num[1:])
        elif num[0] in digits:
            n += 1
            number.map[-n] = int(num[0])
            B(num[1:])
        else:
            raise ValueError(f'parsing failed: invalid input Character "{num[0]}" (position {i}, state A)')
    def B(num):
        global number, i, n
        i += 1
        if not num:
            number.increment_power_by(n)
            return None
        elif num[0] in exponents:
            number.increment_power_by(n)
            D(num[1:])
        elif num[0] in '.':
            number.increment_power_by(n)
            n = 0
            C(num[1:])
        elif num[0] in digits:
            n += 1
            number.map[-n] = int(num[0])
            B(num[1:])
        else:
            raise ValueError(f'parsing failed: invalid Character "{num[0]}" (position {i}, state B)')
    def C(num):
        global number, i, n
        i += 1
        if not num:
            return None
        elif num[0] in exponents:
            D(num[1:])
        elif num[0] in digits:
            n += 1
            number.map[-n] = int(num[0])
            C(num[1:])
        else:
            raise ValueError(f'parsing failed: invalid Character "{num[0]}" (position {i}, state C)')
    def D(num):
        global i, negative_exp, exp
        
        i += 1
        if not num:
            warn('exponent expected but not provided', stacklevel=_warn_stacklevel(4))
            return None
        elif num[0] in '+-':
            if num[0] == '-':
                negative_exp = True
            if not num[1:]:
                warn('exponent expected but not provided', stacklevel=_warn_stacklevel(4))
            exp = 0
            E(num[1:])
        elif num[0] in digits:
            exp = int(num[0])
            E(num[1:])
        else:
            raise ValueError(f'invalid Character "{num[0]}" (position {i}, state D)')
    def E(num):
        global number, i, exp
        i += 1
        if not num:
            sign = 1
            if negative_exp:
                sign = -1
            number.increment_power_by(sign*exp)
            return None
        elif num[0] in digits:
            exp = 10*exp + int(num[0])
            E(num[1:])
        else:
            raise ValueError(f'invalid Character "{num[0]}" (position {i}, state E)')

    A(num)

    p = number.max_power()
    while p in number.map and number.map[p] == 0:
        del number.map[p]
        p -= 1

    if not number.map:
        number.map[0] = 0
        number.zero = True

    return number

def round(*args, **kwargs):
    '''
    round a given number to either
        - a given number of significant figures/significant digits,
        - a given number of decimals, or
        - a given uncertainty
    and optionally output the number with given formatting along with its uncertainty

    Key usage examples:
        round('2.675', sigfigs=2) => '2.7'
        round(2.675, decimals=2)  =>  2.68
        round('2.675', '0.00197') => '2.675 ± 0.002'
        round('123456.789099', '-1.15E-4', output='Drake') => '123 456.789 10(12)

    For detailed usage instructions see https://pypi.org/project/sigfig/
    '''
    if not args:
        warn("no input number given, nothing to return", stacklevel=_warn_stacklevel(2))
        return None
    given = _arguments_parse(args, kwargs)
    num = given['num']

    if num.nan:
        return num.map['NaN']
    if 'decimals' in given:
        num.round_by_decimals(given['decimals'])
    elif 'sigfigs' in given:
        if given['sigfigs'] > len(num.map):
            warn(
                f"{given['sigfigs']} significant figures requested from number with only {len(num.map)} significant figures",
                stacklevel=_warn_stacklevel(2)
            )
        last_power = num.max_power() - given['sigfigs'] + 1
        num.round_by_decimals(-last_power)
        while len(num.map) > given['sigfigs']:
            del num.map[num.min_power()]
    elif 'uncertainty' in given:
        num.has_uncertainty = True
        if 'cutoff' in given:
            cutoff = str(given['cutoff'])
        elif 'cutoff' in _manual_settings:
            cutoff = str(_manual_settings['cutoff'])
        else:
            cutoff = str(_default_settings['cutoff'])
        unc = round(given['uncertainty'], sigfigs=len(cutoff), output='map')
        cut = _num_parse(cutoff + 'E' + str(unc.min_power()))
        if unc > cut:
            unc = round(given['uncertainty'], sigfigs=len(cutoff)-1, output='map')
            if unc.map[unc.max_power()] == 1:
                unc.map[unc.max_power() - 1] = 0
        num.round_by_decimals(-unc.min_power())

    if given['prefix']:
        power_shift = num.prefixify(given['prefix'], given['exponent'])
        if 'uncertainty' in given:
            unc.increment_power_by(power_shift)

    if given['reset_warnings']:
        resetwarnings()

    if issubclass(given['output_type'], (numbers.Number, Decimal)):
        if 'output' in given and given['output'] in {tuple, list}:
            if 'uncertainty' not in given:
                return given['output']([num.output(given['output_type'])])
            return given['output']([num.output(given['output_type']),
                                    unc.output(given['output_type'])])
        return num.output(given['output_type'])
    elif 'output' in given and given['output'] == 'map':
        return num
    
    if 'uncertainty' in given and given['separator'] == 'brackets' and unc.min_power() > 0 and 'external_brackets' not in given:
        return num.decimate(given['format'], unc=unc)

    output = num.decimate(given['format'])
    units = num.prefix if given['prefix'] else ''

    if 'output' in given and given['output'] in {list, tuple}:
        if 'uncertainty' in given:
            return given['output']([output + units, unc.decimate(given['format'], sign=False, units=units)])
        return given['output']([output + units])

    if 'uncertainty' in given:
        if given['separator'] == 'brackets' and unc.min_power() > 0:
            output += '('+unc.decimate(given['format'], sign=False)+')'
        elif given['separator'] == 'brackets':
            output += '('+unc.decimate(given['format'], zeropadding=False, sign=False)+')'
        else:
            if given['prefix'] and given['prefix'] not in {True, 'major', 'minor', 'all'}:
                output += num.prefix
            output += given['separator'] + unc.decimate(given['format'], sign=False)
            if given['prefix'] == True and units:
                output = f'({output})'
    return output + units

def roundit(*args, **kwargs):
    '''Depreciated version of round() function with limited scope'''
    warn('Depreciated Usage: Migrate code to use round() function instead', DeprecationWarning, stacklevel=_warn_stacklevel(2))
    defaults = {'spacer': ' ', 'spacing': 3, 'separator': 'brackets', 'output_type' : str}
    final_parameters = defaults
    if 'form' in kwargs:
        if kwargs['form'] == 'plusminus':
            kwargs['separator'] = ' +/- '
        else:
            kwargs['separator'] = kwargs['form']
        del kwargs['form']
    if 'crop' in kwargs:
        kwargs['crop'] -= 1
    for key in kwargs:
        final_parameters[key] = kwargs[key]
    return round(*args, **final_parameters)
def round_unc(*args, **kwargs):
    '''Depreciated version of round() function with limited scope'''
    warn('Depreciated Usage: Migrate code to use round() function instead', DeprecationWarning, stacklevel=_warn_stacklevel(2))
    defaults = {'sep': tuple}
    final_parameters = defaults
    if 'form' in kwargs and kwargs['form'] == 'plusminus':
        kwargs['separator'] = ' +/- '
        del kwargs['form']
    if 'crop' in kwargs:
        kwargs['crop'] -= 1
    for key in kwargs:
        final_parameters[key] = kwargs[key]
    return round(*[str(arg) for arg in args], **final_parameters)[0]
def round_sf(number, sigfigs):
    '''Depreciated version of round() function with limited scope'''
    warn('Depreciated Usage: Migrate code to use round() function instead', DeprecationWarning, stacklevel=_warn_stacklevel(2))
    return round(str(number),sigfigs=sigfigs)



'''
The SpatialAnalysis class serves a unifying class FOR THE nonGUI API coordinating all the spatial functions/methods in PalmettoBUG 
IT IS NOT USED IN THE GUI AT PRESENT! Instead, the individual spatial classes are called by the GUI. These were made firstm
before the coordinating class, which is why there is a disconnect between the implementation in the GUI and outside it.
'''
# License / derivation info

# This file is licensed under the GPL3 license. No significant portion of the code here is known to be derived from another project 
# (in the sense of needing to be separately / simultaneously licensed)

#The SpatialNeighbors class draws heavily on Squidpy, mainly serving as a way of wrapping squidpy functions in a way that will help them
# play nice with PalmettoBUG's existing data structures 

import os
from typing import Union
from pathlib import Path
import tkinter as tk
import warnings

import tifffile as tf
import skimage
import scipy
import numpy as np
import pandas as pd
import scanpy as sc
import matplotlib.pyplot as plt
import seaborn as sns
import anndata as ann

#from .._vendor.flowsom import FlowSOM, plot_stars

## anticipate de-vendorization:
from flowsom import FlowSOM
from flowsom.pl import plot_stars

from ..Pixel_Classification.Classifiers import smooth_isolated_pixels
from ..Pixel_Classification.use_classifiers import merge_folder
from .SpatialANOVA import SpatialANOVA, plot_spatial_stat_heatmap

warnings.filterwarnings("ignore", message = "Importing read_text")   ## future warning in anndata that squidpy has not caught up to yet. Irritating to see everytime on startup
import squidpy as sq # noqa: E402

__all__ = ["SpatialAnalysis"]

class SpatialAnalysis:
    '''
    This class serves as a coordinating class for the three spatial analysis sub-classes. In the GUI, these subclasses are currently called directly (for historical reasons
    and because there is no real reason to update that). However, for use of PalmettoBUG in scripting outside the GUI, it is convenient to have a unified class where all the 
    spatial methods can be accessed.

    The methods of this class are wrappers on methods of the 3 subclasses. Because of this, it can be divided into a number of groupings:

        >>> add data, cell maps, neighbors, neighborhoods, spaceANOVA, and edt    (in order)

    Args:
        (None) -- the key set-up steps in this class are called in the methods of this class, not when it is initialized

    Key Attributes:
        exp:
            This is the connected Analysis object, containing the anndata (exp.data) which holds most of the data used for calculations

        SpaceANOVA, neighbors, edt:
            These are subclasses that contain the actual methods used by this higher-level class. The methods of this class
            are all wrappers on methods contained in one of these sub-classes. 

    '''
    def __init__(self):
        '''
        '''
        self.edt = SpatialEDT()
        self.SpaceANOVA = SpatialANOVA()
        self.neighbors = SpatialNeighbors()

    def add_Analysis(self, Analysis):
        '''
        Connects the Spatial methods here to a palmettobug.Anlysis object. Edits to the Analysis object will affect the Spatial methods here
        '''
        self.exp = Analysis
        self.edt.add_Analysis(Analysis)
        self.SpaceANOVA.init_analysis(Analysis, output_directory = (self.exp.directory + "/Spatial_plots"))
        self.neighbors.add_Analysis(Analysis)

    def plot_cell_maps(self,
                       plot_type: str,
                       id: Union[str, int, None] = None, 
                       clustering: str = "merging") -> plt.figure:
        '''
        This plot cell maps either as "points" or as "masks".

        Args:
            plot_type (string): 
                either "points" or "masks". If "points" then the cells are represented at dots or various sizes on a white background. If "masks" then 
                the cells are represented as their mask shapes on a black background ("masks" uses a squidpy plotting function). 

            id (string, integer, or None): 
                If None, will make cell maps for every sample_id / image in the dataset. If not None, will only make a plot for the specified image.
                id's should either be the sample_id (for which a string or integer form will work) or the file_name of the desired image.

            clustering (string): 
                the name of the column in self.exp.data.obs to use for coloring the cells. Usually one of the standard clustering column names 
                ("merging", "metaclustering", etc.)

        Returns:
            matplotlib.figure or None
        '''
        if plot_type == "masks":
            if id is None:
                self.neighbors.plot_all_cell_maps(clustering = clustering)
            else:
                id = str(id)
                sample_ids = list(self.exp.data.obs['sample_id'].astype('str'))
                if id in sample_ids:
                    plot, _ = self.neighbors.plot_cell_map(sample_id = id, clustering = clustering, save = True)
                    return plot
                elif id in list(self.exp.data.obs['file_name'].astype('str')):
                    plot, _ = self.neighbors.plot_cell_map(filename = id, clustering = clustering, save = True)
                    return plot
                else:
                    print(f"id = {id} was not in the data.obs sample_id or file_name columns!") 

        elif plot_type == "points":
            if id is None:
                self.SpaceANOVA.plot_cell_maps(multi_or_single = 'ALL', cellType_key = clustering)
            else:
                id = str(id)
                plot = self.SpaceANOVA.plot_cell_maps(multi_or_single = id, cellType_key = clustering)
                return plot

        else:
            print(f"plot_type must be either 'masks' or 'points', but instead {plot_type} was provided") 

    def do_neighbors(self, radius_or_neighbors: str, number: int):
        '''
        Creates the neighbor-graph between cells in the dataset (using their centroids). This step is necessary before performing any of the other neighbor-based methods.
        It uses the squidpy.gr.spatial_neighbors function to generate the neighborhood graph.

        Args:
            radius_or_neighbors (string):
                whether to create the neighbor graph using a fixed radius ("Radius") or to create the neighbor graph using 
                a fixed number of neartest neighbors ("Neighbor")

            number (integer): 
                either the length of the search radius in pixels, or the number of nearest neighbors per cell (depending on radius_or_neighbors parameter)

        Returns:
            (None)
        '''
        self.neighbors.do_neighbors(radius_or_neighbors = radius_or_neighbors, number = number)

    def plot_neighbor_interactions(self, clustering: str = "merging", facet_by: str = "None", col_num: int = 1, filename: Union[str, None] = None):
        '''
        This method wraps squidpy's gr/pl.interaction_matrix functions. It plots a heatmap representing the number of interactions between
        cell types in the dataset. Note that this is an absolute number, and is effected by the abundance of celltypes (more abundant celltypes will 
        have more interactions). 

        Args:
            clustering (string): 
                the name of the column in self.exp.data.obs to use for grouping cells into cell types. Usually one of the standard clustering column names 
                ("merging", "metaclustering", etc.). Their should be >1 unique cluster, the heatmap's dimensions with only one cluster would be 1x1

            facet_by (string): 
                a name of a column in self.exp.data.obs or "None". If not "None", used to interaction matrices for subsets of the data, for example to compare
                interaction matrices between conditions. The first panel is always the interaction matrix for the entire dataset (not subsetted).

            col_num (integer): 
                If facetting, how many columns to have in the figure.

            filename (string or None): 
                if not None, will write the plot as /{filename}.png to the spatial save folder of the palmettobug analysis directory. 

        Returns:
            matplotlib.figure
        '''
        facet_by = str(facet_by)
        plot = self.neighbors.plot_interaction_matrix(clustering = clustering, facet_by = facet_by, col_num = col_num, filename = filename)
        return plot

    def plot_neighbor_enrichment(self, clustering = "merging", facet_by = "None", col_num = 1, seed = 42, n_perms = 1000, filename = None):
        '''
        This method wraps squidpy's gr/pl.neighborhood_enrichment functions. It plots a heatmap representing the enrichment of interactions over the 
        random expectation between cell types in the dataset. This is calculated by permutation test, and the values of the heatmap are z-scores
        between the interactions found in the permutation test and the empirical number of interactions.

        Args:
            clustering (string): 
                the name of the column in self.exp.data.obs to use for grouping cells into cell types. Usually one of the standard clustering column names 
                ("merging", "metaclustering", etc.). Their should be >1 unique cluster, the heatmap's dimensions with only one cluster would be 1x1

            facet_by (string): 
                a name of a column in self.exp.data.obs or "None". If not "None", used to interaction matrices for subsets of the data, for example to compare
                interaction matrices between conditions. The first panel is always the neighborhood enrichment for the entire dataset (not subsetted).

            col_num (integer): 
                If facetting, how many columns to have in the figure.

            seed (integer): 
                random seed for the permutation test

            n_perms (integer): 
                how many permutations to perform in the permutation test

            filename (string or None): 
                if not None, will write the plot as /{filename}.png to the spatial save folder of the palmettobug analysis directory. 

        Returns:
            matplotlib.figure 
        '''
        plot = self.neighbors.plot_neighborhood_enrichment(clustering = clustering, facet_by = facet_by, col_num = col_num, seed = seed, n_perms = n_perms, filename = filename)
        return plot

    def plot_neighbor_centrality(self, clustering: str = "merging", score: str = "closeness_centrality", filename: Union[str,None] = None):
        '''
        Wraps squidpy's gr/pl.centrality_scores functions. clustering corresponds to "cluster_key" in squidpy's API, and score corresponds to "score". 
        
        Args:
           clustering (string): 
                The cell type grouping ('merging', ' metaclustering', etc.) to plot centrality scores for

            score (string):
                The type of centrality score to plot: ['degree_centrality','closeness_centrality','average_clustering']

            filename (string or None):
                If not None, specifies the filename to save the plot under (as a PNG) in the standard /Spatial_plots folder of the analysis folder.
        '''
        plot = self.neighbors.plot_centrality(clustering = clustering, score = score, filename = filename)
        return plot

    def do_neighborhood_CNs(self, 
                            clustering: str = "merging", 
                            leiden_or_flowsom: str  = "FlowSOM", 
                            seed: int  = 42, 
                            resolution: float  = 1.0, 
                            min_dist: float  = 0.1, 
                            n_neighbors: int  = 15, 
                            **kwargs):
        '''
        This method uses a previously constructed neighbor graph and a cell clustering to identify the proportions of each cell type among the neighbors of 
        every cell, then runs an unsupervised clustering algorithm (FlowSOM or Leiden) to group the cells in "cellular neighborhoods" (CNs). This neighborhood 
        grouping is appended to self.exp.data.obs as a "CN" column, and can be used in all the same ways as any other annotation / clustering to generate plots, etc.
        Additionally, a figure is returned that is unique to the type of clustering performed -- if FlowSOM, a minimum spanning tree is returned while if Leiden then
        a UMAP is returned.

        Args:
            clustering (string): 
                the name of the column in self.exp.data.obs to use for grouping cells into cell types. Usually one of the standard clustering column names 
                ("merging", "metaclustering", etc.).

            leiden_or_flowsom (string): 
                "Leiden" or "FlowSOM" -- determines which of the unsupervised clsutering algorithms will be used to group the cells.

            seed (integer): 
                The random seed for the clustering algorithm

            resolution (float): 
                ONLY for Leiden clustering -- corresponds to the same parameter (resolution) in scanpy's tl.leiden function

            min_dist (float): 
                ONLY for Leiden clustering -- corresponds to the same parameter (min_dist) in scanpy's tl.umap function, which is 
                necessary for leiden clustering

            n_neighbors (integer): 
                ONLY for Ledien clustering -- corresponds to the same parameter in scanpy's pp.neighbors function, which is necessary 
                for leiden clustering

            **kwargs: 
                ONLY for FlowSOM -- these passed into the FlowSOM class (copied from saesys lab FlowSOM_Python repository). This allows key
                parameters like the number of trainig cycles (rlen), number of output clusters (n_clusters), and x/y dimensions (xdim and ydim)
                to be passed to the FlowSOM instance. 

        Returns:
            matplotlib.figure
        '''
        figure = self.neighbors.do_cellular_neighborhoods(clustering = clustering, 
                                  leiden_or_flowsom = leiden_or_flowsom, 
                                  seed = seed, 
                                  resolution = resolution, 
                                  min_dist = min_dist, 
                                  n_neighbors = n_neighbors, 
                                  **kwargs)
        return figure

    def plot_CN_graph(self, filename: Union[str, None] = None):
        ''' 
        UMAP or star-plot -- note that this figure is already returned by the method above
        '''
        plot = self.neighbors._plot_stars_CNs(filename = filename)
        return plot

    def plot_CN_heatmap(self, clustering: str = "merging", **kwargs):
        '''
        Plots a heatmap of the proportions of the cell types in each of the CN clusters
        '''
        plot = self.neighbors.plot_CN_heatmap(clustering_col = clustering, **kwargs)
        return plot

    def plot_CN_abundance(self, clustering, cols = 3):
        '''
        Plots a facetted barplot of the proportion of each cell type in each of the CN clusters
        '''
        plot = self.neighbors.plot_CN_abundance(clustering, cols = cols)
        return plot

    def estimate_SpaceANOVA_min_radii(self, with_empty_space = True):
        '''
        This uses information about the cell masks & images (such as perimeter, area, cell occupied bounding-box areas, etc.).

        If with_empty_space is True, will further adjust up the estimating minimum radii using the proportion of empty space in the cell-occupied regions of the images
        '''
        if self.exp.back_up_regions is not None:
            region_props_data = self.exp.back_up_regions
        else:
            region_props_data = self.exp.regionprops_data
        major_r = region_props_data['axis_major_length'] / 2
        minor_r = region_props_data['axis_minor_length'] / 2
        mini = (np.sqrt(major_r * minor_r) / 2).mean()
        if with_empty_space:
            data_table = pd.DataFrame()
            if self.exp.back_up_data is not None:
                data_table['x'] = self.exp.back_up_data.obsm['spatial'][:,0]
                data_table['y'] = self.exp.back_up_data.obsm['spatial'][:,1]
                data_table['sample_id']  = list(self.exp.back_up_data.obs['sample_id'].astype('str'))
            else:
                data_table['x'] = self.exp.data.obsm['spatial'][:,0]
                data_table['y'] = self.exp.data.obsm['spatial'][:,1]
                data_table['sample_id']  = list(self.exp.data.obs['sample_id'].astype('str'))
            minX_Y = data_table[['sample_id','x','y']].groupby('sample_id').min()[['x','y']]
            maxX_Y = data_table[['sample_id','x','y']].groupby('sample_id').max()[['x','y']]
            total_areas = 0
            for i,ii,iii,iv in zip(minX_Y['x'], maxX_Y['x'], minX_Y['y'], maxX_Y['y']):
                total_areas += (ii - i)*(iv - iii)
            empty_space = total_areas - region_props_data['area'].sum()
            mini +=  mini*(empty_space / total_areas)
        return int(mini)

    def do_SpaceANOVA_ripleys_stats(self,
                                    clustering: str,
                                    max: int = 100, 
                                    min: int = 10,
                                    step: int = 1, 
                                    condition1: Union[str, None] = None, 
                                    condition2: Union[str, None] = None,
                                    threshold: int = 10,
                                    permutations: int = 0, 
                                    seed: int = 42, 
                                    center_on_zero: bool = False, 
                                    silence_zero_warnings: bool = True,
                                    suppress_threshold_warnings: bool = False):
        '''
        Calculates Ripley's spatial statistics for every celltype-celltype pair in clustering in every image. The necessary first step in the SpaceANOVA analysis
        pipeline.

        Args:
            clustering (string): 
                the name of the column in self.exp.data.obs to use for grouping cells into cell types. Usually one of the standard clustering column names 
                ("merging", "metaclustering", etc.).

            max / min / step (integers): 
                these are the integers that determine at which radii statistics will be calculated. An easy to see what those radii will be is

                    >>> list(range(min, max, step))

                The default looks at the range 10-100 with a step size of 1. The min is set to be > 0 because the first few radii should essentiall always have
                zero cell interactions in them. This is because we calculate using cell centroids, so even if two cells directly touch, the first few radii will 
                still not have an interaction. For example, if a perfectly circular cell has a diameter of 10, then only radii > 5 will even have a change of encountered 
                another cell since radii < 5 will only search space INSIDE the cell. And even once the earch radius is outside the cell, it does not count as an 
                interaction when it touches another cell -- it only is counted when it touches the centroid of the other cell. 
                Because of this, usually radii < 10 or so have almost no interaction and therefore very unusual behaviour, and might be best dropped from the calculations.
            
            condition1 / condition2 (string or None): 
                If both are None (default), then every condition in the dataset is used (and the fANOVAs are multi-comparison). Else if both
                condition1/2 are specificied, they should be unique values in the 'condition' columns of self.exp.data.obs, and the SpaceANOVA analysis will
                only look at those conditions, using pairwise comparisons / ANOVAs.

            threshold (integer): 
                default = 10. If at least of the celltypes in a given celltype-celltype pair has fewer cells than this threshold in a given image, 
                that image will be skipped & no Ripley statistics will be calculated from that image for that celltype-cetype pair. Note that is a given 
                celltype-celltype pair never passes this threshold for any of the images for a given condition, then it will be ignored for that condition. Further, 
                if only one condition (or no conditions) has images that pass the threshold, then an ANOVA for that celltype-celltype pair is impossible and will be 
                skipped. However, even then, the Ripley's statistics that were successfully calculated for that single condition can still be plotted.

            permutations (integer): 
                If greater than zero, than a permutation correction will be applied to the data. This is done by randomizing the celltype labels in an image
                and calculating the average Ripley's K for those randomizations. The average random K for the celltype is then substracted from the calculated Ripley's K 
                for that celltype in that image. This corrected K is than used to calculate Ripley's L / g as normal.
                Permutation correction can slow the calculation substantially, but is almost always recommended as it uses the actual strucutre of the cells in the 
                images to correct the values of the Ripley's statistics. This is a powerful and simple way to correct for holes / inhomogeneities in the tissue.

            seed (integer): 
                This is the random seed used for the SpaceANOVA methods. This includes the random permutations for the permutation correction, but
                also the seeds used plotting error regions and fANOVA. The seeds for plotting & fANOVA can be set separately when calling those functions
                but by default whatever you use here for the seed will be used for those steps as well. 

            center_on_zero (boolean): 
                ONLY with permutation correction (permutations > 0). This determines whether to 'center' Ripley's g on 0 or on 1.
                Ripley's g is unique among the Ripley's statistics in that it is particularly easy to interpret, as its theoretical value in a 
                random point pattern is equal to a straight line at 1, with values above 1 indicating more association between points that expected,
                and value less than 1 indicating less interactions that expected. However, the permutation correction shifts this centerpoint to 1 when 
                the permutation is substracted from the calculated K. Additionally, when this substract is done the shape of K / L will deviate strongly
                from the theoretical shape of those statistic's curves.
                So: 
                If this parameter is True, then this shift is allowed to occur, and g will need to be interpreted as centered on 0.

                    >>> Permutation correction is: K = K_data - K_permutation

                If this paramter is False, then after substracting the permutaiton K, the theoretical K { pi*(r^2) } is added back, which
                shifts the center of g to 1 without changing its shape. This change also restores the shape of the K / L statistics to better match  
                their more usual, monotonically increasing shape

                    >>> Permutation correction is: K = K_data + (K_theoretical - K_permutation)

            silence_zero_warnings (boolean):
                this method generates a large number of zero division errors, even in a normal run. By default these are silenced. 

            suppress_threshold_warnings (boolean):
                If True, will not print warnings about images failing to meet cell number thresholds
        '''
        self.SpaceANOVA.do_spatial_analysis(
                            condition1 = condition1, 
                            condition2 = condition2,
                            cellType_key = clustering,
                            max = max, 
                            min = min,
                            step = step, 
                            threshold = threshold,
                            permutations = permutations, 
                            seed = seed, 
                            center_on_zero = center_on_zero, 
                            silence_zero_warnings = silence_zero_warnings,
                            suppress_threshold_warnings = suppress_threshold_warnings,
                            )

    def plot_spaceANOVA_function(self, 
                                 stat: str, 
                                 comparison: str = None, 
                                 seed: Union[int, None] = None, 
                                 f_stat: Union[str, None] = None, 
                                 hline: Union[int, None] = None,
                                 output_directory: Union[str, None] = None):
        ''' 
        This function plots a selected Ripley's statistic for a celltype-celltype pair, and optionally also the signle-radii f-values from ANOVA tests
        conducted at each point along the Ripley stat graph.
        
        Args:
            stat (string): 
                either "K", "L", or "g". Determines which Ripley's statistic to plot

            comparison (string or None): 
                a string with the form {celltype1}___{celltype2}. The triple underscore in the middle is how this string is split
                into the two cell types of interest (don't have a triple underscore inside your cell type labels!). A full list of the available comparisons of 
                this form can also be easily accessed with the self.SpaceANOVA._all_comparison_list attribute

            seed (integer or None): 
                the random seed for the fANOVA function. If None, then the seed previously selected in self.do_SpaceANOVA_ripleys_stats will be used.

            f_stat (string or None): 
                if not None, should be "f" (typical), "padj", or "p". If not None, adds a panel to the final plot showing the results of
                (standard, not functional) ANOVA tests comparing conditions at every individual radii. This is useful for visualizing at what distance the 
                difference between conditions is most significant. 

            hline (int or None): 
                if not None, draws a horizontal line on the Ripley's statistics plot at the value (usually ONLY when plotting the 'g' statistic, and 
                set to 0 or 1, depending on where the graph is centered)

            output_directory (string or None): 
                If None, the plots are exported to the automatic / standard directory in the PalmettoBUG project. if not None, should the path to a folder where
                the plots can be exported. (ONLY used if comparison is None)
        '''
        if f_stat is None:
            if comparison is None:
                self.SpaceANOVA.plot_all_spatial(stat = stat, seed = seed, write = True, output_directory = output_directory, hlines = hline)
            else:
                plot = self.SpaceANOVA.plot_spatial(stat_type = stat, 
                                            comparison = comparison, 
                                            seed = seed, 
                                            f_list = f_stat, 
                                            hline = hline)
                plot.savefig(self.SpaceANOVA.output_dir + f"/Functional_plots/{comparison}_{stat}.png", bbox_inches = "tight")
                return plot
        else:
            if comparison is None:
                self.SpaceANOVA.plot_all_spatial_with_stat(seed = seed, 
                                                           stat = stat,
                                                           write = True, 
                                                           output_directory = output_directory, 
                                                           f_p_padj = f_stat, 
                                                           hlines = hline)
            else:
                plot = self.SpaceANOVA.plot_spatial_with_stat(comparison = comparison, 
                                                       seed = seed, 
                                                       stat = stat, 
                                                       f_p_padj = f_stat, 
                                                       hline = hline)
                plot.savefig(self.SpaceANOVA.output_dir + f"/Functional_plots/{comparison}_{stat}.png", bbox_inches = "tight")
                return plot 

    def run_SpaceANOVA_statistics(self, 
                                 stat: str = 'g', 
                                 seed: Union[int, None] = None):
        '''
        This runs the functional ANOVA on the available Ripley's statistics, returning 3 datatables for the (adjusted) p-value, and fANOVA stat
        '''
        self.padj, self.p, self.stat = self.SpaceANOVA.do_all_functional_ANOVAs(stat = stat, seed = seed)
        return self.padj, self.p, self.stat

    def plot_spaceANOVA_heatmap(self, stat: str, filename = None): 
        '''
        Plots a heatmap from one of the dataframes returned / created by self.run_SpaceANOVA_statistics. If plotting a (adjusted) p-value, as is typical, the 
        statistic is transformed by the negative log first so that high number indicate higher significance.

        stat = 'p', 'padj', or 'f'
        '''
        if stat == 'p':
            p_table = self.p
        elif stat == "padj":
            p_table = self.padj
        elif stat == 'f':
            p_table = self.stat
        figure = plot_spatial_stat_heatmap(p_table = p_table, 
                                  vmin = 0, 
                                  vmax = 7)
        
        if filename is not None:
            figure.savefig(self.SpaceANOVA.output_dir + "/" + filename + ".png", bbox_inches = "tight")
        
        return figure

    def do_edt(self, 
               pixel_classifier_folder: str, 
               masks_folder: str, 
               maps: str = "/classification_maps",
               smoothing: int = 10, 
               stat: str = 'mean', 
               normalized: bool = True, 
               background: bool = False,
               marker_class: Union[str, None] = 'spatial_edt', 
               auto_panel: bool = True,
               output_edt_folder: Union[None, str, Path] = None,
               save_path: Union[None, str, Path] = None):
        '''
        Calculates the euclidean distance between cell masks (provided in masks_folder) and matching pixel classifications (pixel_classifier_folder).
        This appends the calculated edt for each cell to self.exp.data, as a new 'channel' / 'antigen', where it can then be used for plotting & calculations.

        Args:
            pixel_classifier_folder (str): 
                the path to a PalmettoBUG-generated pixel classifier folder. This folder needs to contain a subfolder of .tiffs
                containing the pixel class predictions (see maps argument), and contain a biological_labels.csv indicating what the biological
                names of each class in the classifier are.
                When a pixel classifier with > 1 predicted pixel class is used, an edt statistic will be calculated separately for each and added to
                self.exp.data

            masks_folder (str): 
                the path to a folder containing a set of .tiff files of cell segmentation masks. These .tiffs should match those in 
                f'{pixel_classifier_folder}/{maps}'.

            maps (str): 
                should be either "/classification_maps" or "/merged_classification_maps". This determines which subfolder from pixel_classifier_folder
                that contains the .tiff files of the pixel classification maps to usea for calculating the distance from. The filenames of these .tiffs
                should match those in masks_folder

            smoothing (int): 
                If == 0, no smoothing is performed. Otherwise this indicates the size of isolated pixel class regions to smooth out before
                calculating distances. As in, if smoothing == 10 (default) regions of a pixel class smaller than 10 pixels will be dropped and "smoothed"
                into the surrounding pixel classes using mode-based fill-in (the mode of the remaining, closest neighbor pixels will be used to assign the replacement 
                value for dropped pixels). 
                Why smooth? When calculating the distance form a pixel class using the Euclidean Distance Transform (EDT), very small pixel regions can have an outsized
                impact on the final EDT map, so removing spurious / small regions can help clean up the final calculation.

            stat (str): 
                One of "mean", "median", or "minimum". This determines what statistic is read off of each segmentation region when calculating the edt value
                for each cell. The default, "mean", is the most common use, and represents the average distance from the pixel class across the whole cell's spatial
                footprint. "min" means that the calculated value for each cell with just be its minimum distance to the class of interest (its closest point).
                Of note, when "min" is the selected statistic, normalization cannot be performed (see f0ollowing argument). 

            normalized (bool): 
                Whether to normalize (True) or not (False). In this case, normalization means dividing each cell's determined edt value by the average of the 
                image that cell is in. As in, if "median" statistic is selected with normalized = True, each cell's edt value will be:

                    cell_stat = median(cell_edt) / median(image_edt)

                instead of the non-normalized value:

                    cell_stat = median(cell_edt)

                When "mean" is used as a statistic, then the normalization factor is the mean of the image's edt values. Normalization cannot be calculated this way
                with a statistic is "min", and so is ignored (there is no normalization for "min").

                Normalization is useful as it is a way to help correct for the abundance of the pixel class. As in, if one image is 70% within a pixel class (lets say
                the pixel class is for fibrotic regions) its edt distances will be much lower across the image than an image where only 30% of the image is the fibrotic class. 
                This would make all cells - regardless of cell type - in image 1 have much lower edt values than the cells in image 2, which would not be an inaccurate 
                conclusion (the cells in image 1 are genuinely closer to fibrotic regions), but introduces a confounding factor: have the cells moved towards the fibrosis, or 
                has the fibrotic regions expanded?
                Normalization helps address this, in part, as it takes into account the total quantity & positioning of the pixel class in each image when calculating
                the edt value for each cell.

            background (bool): 
                whether to include the 'background' class in the edt calculations (True) or not (False, default). Usually, the background class is ignored,
                because it is not biologically relevant, but in some situations there is effectively no background class / even the background is biologically meaningful.
                For example, if a classifier is trained to identify broad tissue regions in a sample (such as intestinal crypt lumen / epithelia / lamina propia in a 
                colon section), it might be the case that every part of the image falls into one of the pixel classes and there is no 'background'. 
                Because of how supervised classifier's are trained in PalmettoBUG, a background class is always created so this option is useful if that 'background'
                is actually a relevant grouping.

            marker_class (str or None): 
                what marker_class (self.exp.data.var['antigen']) to assign the edt columns when they are added to self.exp.data.X. By default,
                this is "spatial_edt", as it helps the subsequent plottting functions easily find the edt channels while ignoring the other marker_classes.
                However, setting this to "type" / "state" / "none" is also allowed, if you want to perform plots / calculations that combine both the spatial edt
                data and other channels. 

            auto_panel (bool): 
                Whether to automatically add the channels to self.exp.data (True, default), or only to return a panel dataframe (for manually editing
                marker_class, say if you want to assign different edt statistics from a single classifier to different marker_class-es). If False, you will 
                need to manually add the edt data to self.exp.data with self.edt.append_distance_transform(distances_panel = {your edited marker_class panel}).

            output_edt_folder (None, string, or Path): 
                Default is None (no edt map export). If not None, then should be the path where folders of .tiff files can be
                exported. Specifically, the folders used will be the f'{output_edt_folder}_{class_biological_label}' for each class in the classifier.
                The saved .tiff files will be the intermediate Euclidean distances transforms for that class, from which the stats for each cell mask were calculated.

            save_path (None, string, or Path): 
                Default is None (edt valuesa re not saved). If not None, then should be a file path where a csv file can be written.
                This csv will contain the information for all the edt's calculated for the provided pixel classifier / masks pairing. 
                Note that marker_class information will not be saved (that will need to be set again on re-load of the saved edt values)

        Returns:
            a pandas dataframe (panel) which can be used to see or set the marker_class information

            a pandas dataframe (self.edt.results) which contains the edt calculations for every cell and pixel class
        '''
        panel = self.edt.load_distance_transform(pixel_classifier_folder =  pixel_classifier_folder, masks_folder = masks_folder, maps = maps, 
               smoothing = smoothing, stat = stat, normalized = normalized, background = background, marker_class = marker_class, 
               auto_panel = auto_panel, output_edt_folder = output_edt_folder, save_path = save_path)
        return panel, self.edt.results

    def do_reload_edt(self, dataframe, marker_class):
        '''
        Loads a column of data into the anndata of the experiment (meant for saved edt information, but could be used for any type of channel)

        Args:
            dataframe (pandas DataFrame, or Path/string):
                If not a pandas DataFrame, will attempt to pandas.read_csv(dataframe) first. This dataframe should be as long as the number of cells in the data
                and have columns representing spatial_edt data for each cell. Its format should match the format of table exported by do_edt
        '''
        if not isinstance(dataframe, pd.DataFrame):
            dataframe = pd.read_csv(str(dataframe))
        self.edt.reload_edt(dataframe = dataframe, marker_class = marker_class)

    def plot_edt_heatmap(self, groupby_col, marker_class = "spatial_edt", filename = None):
        '''
        Plots a heatmap for spatial edt -- default marker_class is spatial_edt, and export folder (if filename is provided) is in /Spatial_plots

        groupby_col specifies a clustering (such as 'merging', 'metaclustering', etc.) to group the heatmap by
        '''
        if filename is not None:
            plot = self.edt.plot_edt_heatmap(groupby_col = groupby_col, 
                                         marker_class = marker_class, 
                                         filename = self.SpaceANOVA.output_dir + "/" + filename + ".png")
        else:
            plot = self.edt.plot_edt_heatmap(groupby_col = groupby_col, 
                                         marker_class = marker_class, 
                                         filename = None)
        return plot

    def plot_edt_boxplot(self, var_column, groupby_col = 'merging', facet_col = 'condition', col_num = 3, filename = ''):
        '''
        Plots a channel on a horizontal boxplot. Could be used for non-spatial_edt data, but export folder (if a filename is provided)
        is in /Spatial_plots.

        Args:
            var_column (str): 
                the channel to use for the plots. Usually a spatial edt channel (like 'distance to Vimentin'), but could be
                any of the channels in self.exp.data.var['antigen']

            groupby_col (str): 
                the column in self.exp.data.obs that will be used to gruop the box plot (one box per unique value in this column, per facet)

            facet_col (str): 
                the column in self.exp.data.obs that will be used to split the data into two boxplots. Usually (and by default) facetted on 
                the condition column, for comparison of treatment vs. control

            col_num (int): 
                the number of columns of facets before they begin to wrap. As in, with the default of col_num = 3, then the fourth facet
                will be on the second row of the facet grid.

            filename (str): 
                the name to the save a .png file of the plot unde rin /Spatial_plots. If not provided (default value), the plot will not be 
                written to the disk.

        Returns:
            matplotlib.figure (the boxplot)
        '''
        if filename != '':
            plot = self.edt.plot_horizontal_boxplot(var_column = var_column, 
                                                subset_col = groupby_col, 
                                                facet_col = facet_col, 
                                                col_num = col_num, 
                                                filename = self.SpaceANOVA.output_dir + "/" + filename + ".png")
        else:
            plot = self.edt.plot_horizontal_boxplot(var_column = var_column, 
                                                subset_col = groupby_col, 
                                                facet_col = facet_col, 
                                                col_num = col_num, 
                                                filename = '')
        return plot

    def run_edt_statistics(self, 
                           groupby_column: str, 
                           marker_class: str = "spatial_edt", 
                           statistic: str = "mean", 
                           test: str = "anova", 
                           filename: Union[str, None] = None):
        '''
        A wrapper on do_state_exprs from the palmettobug.Analysis class, but with the default marker_class of 'spatial_edt', and a output folder
        in /Spatial_plots.

        Args:
            groupby_column (string): 
                a clustering of the data (such as 'leiden', 'merging', etc.)

            marker_class (string):
                'spatial_edt' (default), 'type','state','none','All' -- if specifying any marker_class except 'spatial_edt', then there is little reason to use
                this function, as you could just use palmettobug.Analysis.do_state_exprs

            statistic (string):
                'mean' or 'median' -- which aggregation method to use when calculating the average value in each ROI / sample_id

            test (string):
                'anova' or 'kruskal' -- whether to use an ANOVA or a Kruskal-Wallis test to do the stats

            filename (string or None): 
                If not None, specifies the filename to save the statistics table under (as a CSV) in the /Spatial_plots folder.

        Returns:
            a pandas dataframe, containing the statistics
        '''
        if filename is not None:
            filename = self.SpaceANOVA.output_dir + "/" + filename + ".csv"
        df = self.edt.plot_edt_statistics(groupby_column = groupby_column, 
                                          marker_class = marker_class, 
                                          statistic = statistic, 
                                          test = test, 
                                          filename = filename)
        return df

class SpatialEDT:
    '''
    
    '''
    def __init__(self):
        pass

    def add_Analysis(self, Analysis_object):
        '''
        '''
        self.exp = Analysis_object

    def load_distance_transform(self, 
                                pixel_classifier_folder: Union[str, None], 
                                masks_folder: Union[str, None], 
                                maps: (str) = "/classification_maps",
                                smoothing: int = 10, 
                                stat: str = 'mean', 
                                normalized: bool = True, 
                                background: bool = False,
                                marker_class: Union[str, None] = 'spatial_edt', 
                                auto_panel: bool = True,
                                output_edt_folder: Union[None, str, Path] = None,
                                save_path: Union[None, str, Path] = None):
        '''
        Loads a distance transform statistic using a folder of masks (should match self.input_mask_folder, if available) and a folder of pixel classifications.
        The distance transform statistic is added for every cell in the dataset & can be treated like an additional antigen -- if marker_class == "type", then
        it can be used in clustering with the rest of the type markers, for example.
        '''
        distance_transforms = pd.DataFrame()   
        class_maps = pixel_classifier_folder + maps
        if marker_class is None:
            marker_class = "state"

        biological_labels = pd.read_csv(pixel_classifier_folder + "/biological_labels.csv")
        if maps == "/merged_classification_maps":
            if not os.path.exists(pixel_classifier_folder + "/merged_classification_maps"):
                merge_folder(pixel_classifier_folder + "/classification_maps",
                            pd.read_csv(pixel_classifier_folder + "/biological_labels.csv"),
                            pixel_classifier_folder + "/merged_classification_maps")
            zipper = zip(biological_labels['labels'].unique(), biological_labels['merging'].unique())
        elif maps == "/classification_maps":
            zipper = zip(biological_labels['labels'], biological_labels['class'])
        else:
            print("ERROR! maps must be either '/classification_maps' or '/merged_classification_maps' ! ")
            raise(ValueError)

        available_file_names = list(self.exp.data.obs['file_name'].unique())

        for i,ii in zipper:
            if ((i != "background") or (background is True)) and (ii != 0):
                if output_edt_folder is not None:
                    output_edt_folder = output_edt_folder + f"_{i}"
                distances = spatial_by_edt_folder(masks_folder, 
                                                  class_maps, 
                                                  available_file_names = available_file_names,
                                                  class_of_interest = ii, 
                                                  stat = stat, 
                                                  normalized = normalized, 
                                                  smoothing = smoothing,
                                                  output_edt_folder = output_edt_folder)
                distance_transforms[f'distance to {i}'] = distances[0]

        if save_path is not None:
            distance_transforms.to_csv(str(save_path), index = False)
        self.results = distance_transforms
        # print(self.results)
        # print(self.exp.data.obs)

        if len(distance_transforms) != len(self.exp.data.obs):
            if self.exp._in_gui: 
                tk.messagebox.showwarning("Warning!", message = "Distance edt data and currently loaded data do not match in length!"
                                           "\nAborting distances load. Did you use the same masks folder for distances calculations and for segmentation, \n"
                                           "Or have you edited that masks folder since starting this experiment (creating a mismatch in the number of \n"
                                           "masks in that folder and number cell events in the loaded analysis)?")
            else:
                print("Distance edt data and currently loaded data do not match in length!"
                      "\nAborting distances load. Did you use the same masks folder for distances calculations and for segmentation, \n"
                      "Or have you edited that masks folder since starting this experiment (creating a mismatch in the number of \n"
                      "masks in that folder and number cell events in the loaded analysis)?")
            return
        self.exp._distance_edt_data = distance_transforms
        if self.exp._in_gui: 
            self.exp.logger.info(f"Performed first step of adding distance transform edt data"
                            f"\n pixel_classifier = {pixel_classifier_folder}, masks = {masks_folder}, smoothing = {str(smoothing)}") 
            
        distances_panel = pd.DataFrame()
        distances_panel['fcs_colnames'] = distance_transforms.columns
        distances_panel['antigen'] = distance_transforms.columns
        distances_panel['marker_class'] = marker_class
        distances_panel.to_csv(self.exp.directory + '/distances_edt_panel.csv', index = False) 
        if auto_panel:
            self.append_distance_transform(distances_panel = distances_panel) ## specifying distances panel in this call overwrites any prior panel
        return distances_panel
    
    def append_distance_transform(self, distances_panel: Union[str, pd.DataFrame, None] = None):
        ''' 
        Adds a distance transform statistic loaded by load_distance_transform to self.data so that it can be accessed.
        '''
        if distances_panel is None:
            self.distances_panel = pd.read_csv(self.exp.directory + '/distances_edt_panel.csv')
        else:
            if isinstance(distances_panel, pd.DataFrame):
                self.distances_panel = distances_panel
            else:
                self.distances_panel = pd.read_csv(str(distances_panel))

        self.distances_panel.index = self.distances_panel['antigen']

        for i in self.exp.data.var['antigen']:         ## prevent duplicate columns by removing any prior columns/antigens that match any of the names of the new columns being added
            if i in self.exp._distance_edt_data.columns:
                slicer = self.exp.data.var['antigen'] != i
                self.exp.data = self.exp.data[:, slicer]
                self.exp.panel = self.exp.panel[slicer]

        self.exp.panel = pd.concat([self.exp.panel, self.distances_panel], axis = 0)

        if self.exp.unscaled_data is None:
            new_X  = np.concatenate((self.exp.data.X.copy(), np.array(self.exp._distance_edt_data)), axis = 1)
            self.exp.data = ann.AnnData(X = new_X, 
                                        var = self.exp.panel, 
                                        varm = self.exp.data.varm, 
                                        varp = self.exp.data.varp,
                                        obs = self.exp.data.obs, 
                                        obsm = self.exp.data.obsm, 
                                        obsp = self.exp.data.obsp, 
                                        uns = self.exp.data.uns)
        else:
            new_X  = np.concatenate((self.exp.unscaled_data.copy(), np.array(self.exp._distance_edt_data)), axis = 1)
            self.exp.unscaled_data = new_X.copy()
            self.exp.data = ann.AnnData(X = new_X, 
                                        var = self.exp.panel, 
                                        varm = self.exp.data.varm, 
                                        varp = self.exp.data.varp,
                                        obs = self.exp.data.obs, 
                                        obsm = self.exp.data.obsm, 
                                        obsp = self.exp.data.obsp, 
                                        uns = self.exp.data.uns)
            self.exp.do_scaling(scaling_algorithm = self.exp._scaling, upper_quantile = self.exp._quantile_choice)       

        if self.exp._in_gui: 
            self.exp.logger.info("Appended distance transform edt data") 

    def reload_edt(self, dataframe, marker_class):
        '''
        '''
        self.exp._distance_edt_data = dataframe
        distances_panel = pd.DataFrame()
        distances_panel['fcs_colnames'] = dataframe.columns
        distances_panel['antigen'] = dataframe.columns
        distances_panel['marker_class'] = marker_class

        if len(dataframe) != len(self.exp.data.obs):
            if self.exp._in_gui: 
                tk.messagebox.showwarning("Warning!", message = "Distance edt data and currently loaded data do not match in length!"
                                           "\nAborting distances load. Did you use the same masks folder for distances calculations and for segmentation, \n"
                                           "Or have you edited that masks folder since starting this experiment (creating a mismatch in the number of \n"
                                           "masks in that folder and number cel events in the loaded analysis)?")
            else:
                print("Distance edt data and currently loaded data do not match in length!"
                      "\nAborting distances load. Did you use the same masks folder for distances calculations and for segmentation, \n"
                      "Or have you edited that masks folder since starting this experiment (creating a mismatch in the number of \n"
                      "masks in that folder and number cell events in the loaded analysis)?")
            return
        self.append_distance_transform(distances_panel = distances_panel) ## specifying distances panel in this call overwrites any prior panel
        return distances_panel

    def plot_horizontal_boxplot(self,
                                var_column: str, 
                                subset_col: str = 'merging', 
                                facet_col: str = 'condition', 
                                col_num: int = 3, 
                                filename: str = ''):
        '''
        This function is for plotting the dstirubtion of values in a single column in an anndata object.
    
        Args:
            var_column (string): a unique value in anndata.var['antigen'] that denotes whivh column of anndata.X to use

            subset_col (string):  the name of a column in anndata.obs that will be used to subset the data into columns

            facet_col (string): the name of a column in anndata.obs that will be used to facet the plot into multiple panel (usually by experimental condition)

            col_num (integer): the number of columns of the plot panels

            filename (string): if not '', the FULL filepatth where the plot will be saved 
    
        Returns:
            maptlotlib.pyplot figure (the horizontal boxplot)
    
        I/O:
            Outputs: a file at the full filepath [filename], this is the saved version of the returned matplotlib.pyplot figure -- it is usually a .png file.
        '''
        anndata = self.exp.data.copy()
        df = anndata.obs
        df[var_column] = list(np.squeeze(anndata.X[:,anndata.var['antigen'] == var_column]))
        
        facet_num = len(df[facet_col].unique())
        if facet_num == col_num:
            col_num -= 1
    
        row_num = (facet_num // col_num) + 1 
        if (facet_num % col_num) == 0:
            row_num -= 1
        
        figure, axs = plt.subplots(row_num, col_num, sharey = True, sharex = True, figsize = (11,11))
        
        for ii,i in enumerate(df[facet_col].unique()):
            to_plot = df[df[facet_col] == i]
            to_plot.loc[:,subset_col] = to_plot[subset_col].astype('str')
            ax = axs.ravel()[ii]
            if subset_col != "None":
                sns.boxplot(to_plot, x = var_column, y = subset_col, orient = "h", ax = ax)
            else:
                sns.boxplot(to_plot, x = var_column, orient = "h", ax = ax)
            ax.set_title(i)
            
        ii += 1
        while (len(axs.ravel()) - ii) > 0:
            axs.ravel()[ii].set_axis_off()
            ii += 1
            
        if (filename != '') and (filename is not None):
            figure.savefig(filename, bbox_inches = 'tight')
        plt.close()
        return figure


    def plot_edt_heatmap(self, groupby_col, marker_class = "spatial_edt", filename = None):
        '''
        This is just a wrapper on pbug.Analysis.plot_medians_heatmap, where the default marker_class is "spatial_edt"
        '''
        figure = self.exp.plot_medians_heatmap(marker_class = marker_class, groupby = groupby_col)
        if filename is not None:
            figure.savefig(filename, bbox_inches = "tight")
        plt.close()
        return figure


    def plot_edt_statistics(self, groupby_column, marker_class = "spatial_edt", statistic = "mean", test = "anova", filename = None):
        '''
        This is just a wrapper on pbug.Analysis.do_state_exprs_ANOVAs, where the default marker_class is "spatial_edt"
        '''
        stat_df = self.exp.do_state_exprs_ANOVAs(groupby_column = groupby_column, 
                                                 marker_class = marker_class,
                                                 statistic = statistic,
                                                 test = test)
        if filename is not None:
            stat_df.to_csv(filename, index = False)
        return stat_df

def spatial_by_edt_folder(masks_folder: Union[str, Path], 
                          class_map_folder: Union[str, Path], 
                          available_file_names = None,
                          class_of_interest: int = 2, 
                          do_all_classes: int = 0, 
                          stat: str  = "mean", 
                          normalized: bool  = True, 
                          smoothing: int  = 0,
                          output_edt_folder: Union[None, str, Path] = None) -> list[list[float]]:
    '''
    from a folder of cell masks and amatching folder of pixel classifications, calculates a distance transform statistic for every cell in the data.
    This list of distance transform statistics can then be easily added to an anndata object containing read outs from the matching cells.

    Args:
        masks_folder (string or Path): the filepath to a folder containing the masks to be used in the calculation

        class_map_folder (string or Path): the filepath to a folder containing the pixel classifications to be used in the calculation

        class_of_interest (integer): (only if do_all_classes == 0) which class to calculate distance transform statistics for

        do_all_classes (integer): either 0 (when only the class_of_interest will be used) or a number indicating how many pixel classes to use from the 
                classifier (the first X classes will have distance transform stats calculated where X = do_all_classes)

        stat (integer): which statistic to calculate -- "mean" will calculate the mean euclidean distance transform value in each mask
                "median" will calculate the median value, and "min" the minimum. 

        normalized (boolean): whether to normalize the results of each mask. Only if stat != "min". As in, instead of reporting the raw
                average euclidean distance transform value within each mask, will report that value divided by the average of the image 
                that mask was in 

                        --> mean(edt_of_mask) / mean(edt_of_image) -- or median / median as the case may be

                this helps reduce the effect of some images containing substantially more of the pixel class

        smoothing (integer): a number indicating a threshold. Isolated regions of pixel class smaller than this number will be "smoothed" out (dropped to 
                zero, then their values replaced by the mode of the remaining, surrounding pixels). Effectively removes small regions in the pixel classification
                with the goal of removing spurious classifications. Particularly can be important for these distance transforms, as even a single isolated pixel in the 
                class of interest can create a large circular region of lower edt values.
                If no smoothing is desired, set this parameter == 0.

        output_edt_folder (None, string, or Path): If not None, then should be a file path to a folder where the edt maps can be exported. If None, edt maps are not exported.

    Returns:
        a list of lists of floats -- each sublist is the spatial transform for the full data for one of the pixel classes in the selected classifier
                if only one class is selected, then this will be a list containing a single sublist -- which will be as long as the data set (one float / distance
                statistic per cell)
    '''
    masks_folder = str(masks_folder)
    class_map_folder = str(class_map_folder)
    if output_edt_folder is not None:
        if not os.path.exists(output_edt_folder):
            os.mkdir(output_edt_folder)
    edt_list = []
    masks = [masks_folder + "/" + i for i in sorted(os.listdir(masks_folder)) if i.lower().find(".tif") != -1]
    class_maps = [class_map_folder + "/" + i for i in sorted(os.listdir(class_map_folder)) if i.lower().find(".tif") != -1]
    #print(available_file_names)
    if available_file_names is not None:
        mask_short = [i for i in sorted(os.listdir(masks_folder)) if i.lower().find(".tif") != -1]
        masks = [i for i,ii in zip(masks, mask_short) if f"{ii[:ii.rfind('.')]}.fcs" in available_file_names]
        maps_short = [i for i in sorted(os.listdir(class_map_folder)) if i.lower().find(".tif") != -1]
        class_maps = [i for i,ii in zip(class_maps, maps_short) if f"{ii[:ii.rfind('.')]}.fcs" in available_file_names]     
        
    for iii, (i,ii) in enumerate(zip(masks, class_maps)):
        mask = tf.imread(i).astype('int')
        class_map = tf.imread(ii).astype('int')
        if output_edt_folder is not None:
            output_edt_path = output_edt_folder + f"/{i[i.rfind('/'):]}"
        else:
            output_edt_path = None
        if do_all_classes != 0:
            for i in range(0,do_all_classes,1):
                if iii == 0:
                    edt_list.append([])
                edt_list[iii] = edt_list[iii] + _spatial_by_edt(mask, class_map, 
                                                                class_of_interest = i, 
                                                                stat = stat, 
                                                                normalized = normalized, 
                                                                smoothing = smoothing, 
                                                                output_edt_path = output_edt_path)
        else:
            if iii == 0:
                edt_list.append([])
            edt_list[0] = edt_list[0] + _spatial_by_edt(mask, 
                                                        class_map, 
                                                        class_of_interest, 
                                                        stat = stat, 
                                                        normalized = normalized, 
                                                        smoothing = smoothing, 
                                                        output_edt_path = output_edt_path)
    return edt_list

def _spatial_by_edt(mask: np.ndarray[int], 
                    class_map: np.ndarray[int], 
                    class_of_interest: int, 
                    stat: str = 'mean', 
                    normalized: bool = True, 
                    smoothing: int = 0,
                    output_edt_path: Union[None, str, Path] = None):
    '''
    helper for spatial_by_edt_folder. From the provided numpy arrays representing the pixel classification & mask, calculates a distance transform statistic for each cell
    '''
    edt_list = []
    if smoothing > 0:
        max = np.max(class_map)
        #class_map[class_map == 0] = max + 1
        class_map = smooth_isolated_pixels(class_map, 
                                           class_num = max, 
                                           threshold = smoothing, 
                                           mode_mode = "dropped_image", 
                                           fill_in = False,
                                           warn = False)
        #class_map[class_map == max + 1] = 0
    class_map = (class_map != class_of_interest).astype('int') 
    edt_map = scipy.ndimage.distance_transform_edt(class_map)
    if output_edt_path is not None:
        output_edt_path = str(output_edt_path)
        tf.imwrite(output_edt_path, edt_map)
        
    if (stat == "mean") and (normalized is True):
        for j in skimage.measure.regionprops(mask, edt_map):
            edt_list.append(np.nan_to_num(np.asarray(j.intensity_mean / np.mean(edt_map)), 0))
    elif (stat =="mean") and (normalized is False):
        for j in skimage.measure.regionprops(mask, edt_map):
            edt_list.append(j.intensity_mean)

    elif (stat == "median") and (normalized is False):
        for j in skimage.measure.regionprops(mask, edt_map):
            edt_list.append(np.median(j.image_intensity))
    elif (stat == "median") and (normalized is True):
        for j in skimage.measure.regionprops(mask, edt_map):
            edt_list.append(np.nan_to_num(np.asarray(np.median(j.image_intensity) / np.median(edt_map)), 0))

    elif stat == "min":
        for j in skimage.measure.regionprops(mask, edt_map):
            edt_list.append(j.intensity_min)

    return edt_list


class SpatialNeighbors:        ## formerly SquipySpatial
    '''
    This class coordinates primarily squidpy-based neighborhood analysis options, such as neighborhood enrichment, cellular neighborhood groupings, etc.

    Key Attributes:
        self.exp (PalmettoBUG.Analysis): the analysis object containing the data 

        self.exp.data (anndata): the primary anndata object on which scanpy and squidpy operations are performed. It tends to get edited-in-place by many 
                of the methods, which can be useful when using an anndata object shared with an instance of pbug.Analysis(.data)

        self.save_dir, self.save_cell_maps_dir (string): paths to standard locations where plots are / can be saved

        self.masks_paths (list of strings): a list of full filepaths to the mask file that corespond to the data in self.data. For example, self.data.obs['file_name'] should 
                contain matching filenames (but not FULL file paths) to those in this list.

    '''
    def __init__(self):
        self.data = None

    def add_Analysis(self, Analysis):
        ''''''
        self.exp = Analysis
        masks_path = self.exp.input_mask_folder
        save_directory = (self.exp.directory + "/Spatial_plots") 
        self.save_dir = str(save_directory)
        if not os.path.exists(self.save_dir):
            os.mkdir(self.save_dir)
        self.save_cell_maps_dir = self.save_dir + "/cell_maps"
        if not os.path.exists(self.save_cell_maps_dir):
            os.mkdir(self.save_cell_maps_dir)
        self.masks_paths = [masks_path + "/" + i for i in os.listdir(str(masks_path)) if i.lower().find(".tif") != -1]

    def do_neighbors(self, radius_or_neighbors: str, number: int):
        '''
        Wrapper for squidpy.gr.spatial_neighbors --> preformed on self.exp.data using the provided radius and n_neighbors
        '''
        if radius_or_neighbors == "Radius":
            sq.gr.spatial_neighbors(self.exp.data, coord_type = "generic", library_key = "sample_id", radius = number)
        elif radius_or_neighbors == "Neighbors":
            sq.gr.spatial_neighbors(self.exp.data, coord_type = "generic", library_key = "sample_id", n_neighs = number)
        else:
            print(f"radius_or_neighbors must be either 'Radius' or 'Neighbors', but instead was {radius_or_neighbors}")

    def do_cellular_neighborhoods(self, 
                                  clustering: str = "merging", 
                                  leiden_or_flowsom: str  = "FlowSOM", 
                                  n_clusters = 20,
                                  seed: int  = 42, 
                                  resolution: float  = 1.0, 
                                  min_dist: float  = 0.1, 
                                  n_neighbors: int  = 15,
                                  flavor = "leidenalg", 
                                  plot_filename: Union[None, str] = None,
                                  **kwargs):
        '''
        Performs either FlowSOM-based or Leiden-based clustering of cells into cellular neighborhoods.

        Args:
            clustering (str): the cell metadata column (in self.exp.data.obs) used as input for the chosen clustering algorithm. As in, every cells' neighbors for each 
                    category in this metadata column is counted and a % is derived. This % neighboring cell types is tehn passed into the clustering
                    algorithm.

            leiden_or_flowsom (str): One of two strings -- "FlowSOM" or "Leiden" -- determines which algorith m is used to cluster the cells

            n_clusters (int): FLOWSOM only! -- the number of metaclustering to return at the end of the algorithm

            seed (int): random seed for reproducibility

            resolution (float): (leiden only) the resolution of the leiden clustering. High numbers tend to increase the number of cluster found
                    while lower numbers reduce the number of clusters found

            min_dist (float):  (leiden only) the minimum distance for the UMAP performed alongside the leiden clustering (I think this does not effect the leiden, but
                    does affect the UMAP secondary output.)

            n_neighbors (integer): (leiden only) this is used in scanpy.pp.neighbors() call of the %neighboring cell types data, which precedes leiden clustering

            **kwargs: (FlowSOM only) these keyword arguments are passed into the FlowSOM algorithm. arguments like rlen, etc. are commonly used.

        Returns:
            matplotlib.pyplot figure (UMAP for Leiden clustering, Minimum Spanning Tree / "starplot" for FlowSOM)

        Edits-in-place:
            self.exp.data: adds a 'CN' column in self.exp.data.obs, with the calculated cellular neighborhood groupings. 
        '''
        ### requires self.do_neighbors() first
        unique_clusters = self.exp.data.obs[clustering].unique()
        special_category = pd.CategoricalDtype(categories = unique_clusters, ordered = True)
        my_anndata = self.exp.data.copy()
        final_array = np.zeros([len(my_anndata), len(unique_clusters)])
        offset = 0
        for j in my_anndata.obs['sample_id'].unique():
            temp_anndata = my_anndata[my_anndata.obs['sample_id'] == j].copy()
            dense = temp_anndata.obsp['spatial_connectivities'].todense()
            for ii,i in enumerate(dense):
                cell_neighbor_types = pd.DataFrame(temp_anndata.obs.loc[np.array(i == 1).squeeze(), clustering].astype(special_category)).reset_index()
                percentages = np.array(cell_neighbor_types.groupby(clustering, observed = False).count() / len(cell_neighbor_types))
                final_array[ii + offset,:] = percentages.squeeze()
            offset += len(dense)
        final_array = np.nan_to_num(final_array, 0)
        var = pd.DataFrame()
        var['antigen'] = my_anndata.obs[clustering].unique()
        var.index = var['antigen']
        new_anndata = ann.AnnData(final_array, var = var)
        if leiden_or_flowsom.lower() == "flowsom":
            self.neighbors_flowsom = FlowSOM(new_anndata, seed = seed, n_clusters = n_clusters, **kwargs) 
            self.exp.data.obs['CN'] = list(self.neighbors_flowsom.get_cell_data().obs['metaclustering'])
            self.exp.data.obs['CN'] = self.exp.data.obs['CN'].astype('int') + 1
            figure = self._plot_stars_CNs()
            figure.subplots_adjust(top = 1, left = 0, bottom = 0)
            figure.suptitle("Cell Neighborhood FlowSOM Minimum Spanning Tree")
        

        elif leiden_or_flowsom.lower() == "leiden":
            rn_gen = np.random.default_rng(seed = seed)
            new_anndata.X = new_anndata.X + (rn_gen.random(new_anndata.X.shape) / 100)     ## Mind the eigengap
            sc.pp.neighbors(new_anndata, n_neighbors = n_neighbors, random_state = seed)
            sc.tl.umap(new_anndata, 
                        min_dist = min_dist, 
                        random_state = seed)
            sc.tl.leiden(new_anndata, 
                        resolution = resolution, 
                        random_state = seed,
                        flavor = flavor, 
                        n_iterations = 2)
            new_anndata.obs['leiden'] = new_anndata.obs['leiden'].astype('int') + 1
            new_anndata.obs['leiden'] = new_anndata.obs['leiden'].astype('category')
            self.exp.data.obs['CN'] = list(new_anndata.obs['leiden'])
            figure = plt.figure()
            ax = plt.gca()
            sc.pl.umap(new_anndata, color = "leiden", show = False, ax = ax)
            figure.suptitle("Leiden clustering of neighborhoods defined by % cell types")
            plt.close()
        if plot_filename is not None:
            figure.savefig(self.save_dir + "/" + plot_filename + ".png", bbox_inches = "tight") 
        return figure
        
    def plot_CN_heatmap(self, clustering_col: str = "merging", **kwargs):
        '''
        Plots a heatmap of the percent of {clustering_col} cells in each cellular neighborhood (CN). Useful for identifying what each CN is
        '''
        data = self.exp.data.copy()
        data.obs['CN'] = data.obs['CN'].astype('category')
        towards_percents = data.obs.groupby(['CN', clustering_col], observed = False).count().reset_index()
        cluster_df = pd.DataFrame()
        for ii,i in enumerate(towards_percents['CN'].unique()):
            subset = towards_percents[towards_percents['CN'] == i]
            if ii == 0:
                cluster_df[clustering_col] = subset[clustering_col]
            cluster_df[i] = list(subset['sample_id'] / subset['sample_id'].sum())

        cluster_df.index = cluster_df[clustering_col]

        def min_max(array):
            return (array - array.min()) / (array.max() - array.min())
        
        to_plot_df = cluster_df.drop(clustering_col, axis = 1).apply(min_max, axis = 1)
        to_plot_df = to_plot_df.T.reset_index()
        to_plot_df['CN cluster'] = to_plot_df['index']
        to_plot_df.index = to_plot_df['CN cluster']
        to_plot_df = to_plot_df.drop(['CN cluster', 'index'], axis = 1)
        figure = sns.clustermap(to_plot_df, xticklabels = 1, yticklabels = 1, **kwargs)
        figure.figure.suptitle('Scaled Percent of each celltype within each Cellular Neighborhood cluster')
        plt.close()
        return figure.figure

    def plot_CN_abundance(self, clustering_col, cols = 3):
        '''
        Plot the abundance of the celltypes (determined by clustering_col) in each cellular neighborhood
        '''
        data = self.exp.data.copy()
        data.obs['CN'] = data.obs['CN'].astype('category')
        towards_abundance = data.obs.groupby(['CN', clustering_col], observed = False).count().reset_index()
        facets = len(towards_abundance['CN'].unique())
        if int(facets) == int(cols):
            cols -= 1    ## automatically reshape if it is too small

        rows = (facets // cols) + 1            ## + 1 because the // operation rounds down
        if (rows == 1) and (cols > 2):
            cols -= 1  
            rows = (facets // cols) + 1  

        if int(facets % cols) == 0:
            rows = rows - 1   ## avoid a blank row at the end
        figX = cols * 2.35
        figY = rows * 1.9
        figure, axs = plt.subplots(rows, 
                                   cols, 
                                   sharex = True, 
                                   sharey = True, 
                                   figsize = (figX, figY))
        axs = axs.ravel()
        for i,ii in enumerate(towards_abundance['CN'].unique()):
            slice_df = towards_abundance[towards_abundance['CN'] == ii].copy()
            slice_df['% in neighborhood'] = slice_df['sample_id']
            sns.barplot(x = slice_df[clustering_col], 
                        y = (slice_df['% in neighborhood'] / slice_df['sample_id'].sum()) * 100, ### just need a column that consistently is in self.exp.data.obs
                        ax = axs[i])   
            axs[i].set_title(str(ii))
            axs[i].tick_params("x", labelrotation = 90)
        
        i += 1
        while len(axs) > i:
            axs[i].set_axis_off()
            i += 1

        figure.suptitle('Percentage of each celltype in the cellular Neighborhood clusterings')
            
        return figure

    def _plot_stars_CNs(self, filename: Union[str, None] = None):
        '''
        Plots the minimum spanning tree / star plot from the FlowSOM package
        '''
        figure = plot_stars(self.neighbors_flowsom, 
                         markers=None, 
                         background_values = self.neighbors_flowsom.get_cluster_data().obs['metaclustering'], 
                         title=None)
        figure.set_size_inches(18,18)
        sns.move_legend(figure.axes[0], loc = 'lower right')
        if filename is not None:
            figure.savefig(self.save_dir + "/" + filename, bbox_inches = "tight") 
        plt.close()
        return figure

    def plot_neighborhood_enrichment(self, clustering = "merging", facet_by = "condition", col_num = 1, seed = 42, n_perms = 1000, filename = None):
        '''
        Plots neighborhood enrichment scores for the chosen [clustering]. Can be facetted using [facet_by] to compare things like condition
        '''
        data = self.exp.data.copy()
        data.obs[clustering] = data.obs[clustering].astype('category')
        sq.gr.nhood_enrichment(data, cluster_key = clustering, show_progress_bar = False, seed = seed, n_perms = n_perms)
        if facet_by != "None":
            conditions = data.obs[facet_by].unique()
            condition_num = len(conditions)
            if condition_num < 2:
                facet_by = None
                figure = plt.figure()
                ax = plt.gca()
            else:
                if (condition_num + 1) % col_num > 0:
                    row_num = ((condition_num + 1) // col_num) + 1
                else:
                    row_num = (condition_num + 1) // col_num
                figure, axs = plt.subplots(row_num, col_num, sharey = True, figsize = (col_num * 6, row_num * 4))
                axs = axs.ravel()
                ax = axs[0]
        else:
            figure = plt.figure()
            ax = plt.gca()
        sq.pl.nhood_enrichment(data, cluster_key = clustering, cmap = 'coolwarm', ax = ax, fontsize = 10)

        if facet_by != "None":
            for ii,i in enumerate(conditions):
                ax = axs[ii + 1]
                condition_anndata = data[data.obs[facet_by] == i].copy()
                sq.gr.nhood_enrichment(condition_anndata, cluster_key = clustering, show_progress_bar = False, seed = seed, n_perms = n_perms)
                sq.pl.nhood_enrichment(condition_anndata, 
                                       cluster_key = clustering, 
                                       cmap = 'coolwarm', 
                                       ax = ax, 
                                       fontsize = 10, 
                                       title = f'{i} Neighborhood Enrichment')
            diff = len(axs) - (ii + 1)
            for j in range(0, diff + 1):
                axs[ii + j].set_axis_off()
                
        if filename is not None:
            figure.savefig(self.save_dir + "/" + filename, bbox_inches = "tight") 
        plt.close()
        return figure
    
    def plot_interaction_matrix(self, clustering: str = "merging", facet_by: str = "None", col_num: int = 1, filename: Union[str, None] = None):
        '''
        Plots the ineteraction matrix for the chosen [clustering]. Can be facetted using [facet_by] to compare things like condition.

        Neighborhood enrichment is typically better / more informative as that represents interactions beyond expectation, while the interaction matrix
        plotted by this function is just the total count of interactions (so abundant cell types are likely to have higher interaction numbers than the other
        cell types), and does not carry any statistical significance on its own.)
        '''
        data = self.exp.data.copy()
        data.obs[clustering] = data.obs[clustering].astype('category')
        sq.gr.interaction_matrix(data, cluster_key = clustering)
        if facet_by != "None":
            conditions = data.obs[facet_by].unique()
            condition_num = len(conditions)
            if condition_num < 2:
                facet_by = None
                figure = plt.figure()
                ax = plt.gca()
            else:
                if (condition_num + 1) % col_num > 0:
                    row_num = ((condition_num + 1) // col_num) + 1
                else:
                    row_num = (condition_num + 1) // col_num
                figure, axs = plt.subplots(row_num, col_num, sharey = True)
                axs = axs.ravel()
                ax = axs[0]
        else:
            figure = plt.figure()
            ax = plt.gca()
        sq.pl.interaction_matrix(data, cluster_key = clustering, cmap = 'coolwarm', ax = ax)

        if facet_by != "None":
            for ii,i in enumerate(conditions):
                ax = axs[ii + 1]
                condition_anndata = data[data.obs[facet_by] == i].copy()
                sq.gr.interaction_matrix(condition_anndata, cluster_key = clustering)
                sq.pl.interaction_matrix(condition_anndata, 
                                         cluster_key = clustering, 
                                         cmap = 'coolwarm', 
                                         ax = ax, 
                                         fontsize = 10, 
                                         title = f'{i} Interaction Matrix')
            diff = len(axs) - (ii + 1)
            for j in range(0, diff + 1):
                axs[ii + j].set_axis_off()
                
        if filename is not None:
            figure.savefig(self.save_dir + "/" + filename, bbox_inches = "tight") 
        plt.close()
        return figure
    
    def plot_centrality(self, clustering: str = "merging", score: str = "closeness_centrality", filename: Union[str,None] = None):
        '''
        score_options = ["closeness_centrality", "degree_centrality", "average_clustering"]

        See squipy's documentation on squidpy.gr.centrality_scores() for what these represent

        Because of how squipdy.pl.centrality() works, cannot be facetted.
        '''  
        data = self.exp.data.copy()
        data.obs[clustering] = data.obs[clustering].astype('category')
        sq.gr.centrality_scores(data, cluster_key = clustering, score = score, show_progress_bar = False)
        if filename is not None:
            sq.pl.centrality_scores(data, cluster_key = clustering, score = score, save = self.save_dir + "/" + filename)
        else:
            sq.pl.centrality_scores(data, cluster_key = clustering, score = score)
        figure = plt.gcf()
        plt.close()
        return figure
    
    def plot_cell_map(self, 
                      sample_id: Union[str,None] = None, 
                      filename: Union[str,None] = None, 
                      clustering: str = "merging", 
                      save: bool = False):
        '''
        Plots a single image (designated by either sample_id or filename) in the style of squidpy.pl.spatial_segment (masks shapes plotted, colored by [clustering])
        '''
        if (filename is None) and (sample_id is not None):
            filename = self.exp.data.obs[self.exp.data.obs['sample_id'] == sample_id]['file_name'].values[0]

        if self.exp.back_up_data is not None:
            data = self.exp.back_up_data.copy()
            up_to_date_data = self.exp.data.copy()
            data.obs[clustering] = 'dropped'
            data.obs.loc[up_to_date_data.obs.index, clustering] = up_to_date_data.obs[clustering]
        else:
            data = self.exp.data.copy()

        unique_clusters = data.obs[clustering].unique()
        special_category = pd.CategoricalDtype(categories = unique_clusters, ordered = True)
        subset_anndata = data[data.obs['file_name'] == filename].copy()
        subset_anndata.obs[clustering] = subset_anndata.obs[clustering].astype(special_category)
        filename = filename[:filename.rfind(".")]
        mask_path = [i for i in self.masks_paths if (i.find(filename) != - 1)]
        if len(mask_path) > 1:                 ## if a given filename is fully present within another filename in the folder (so .find() != -1 for >1 of the paths), 
                                                            # always take the filepath with theshortest match
            mask_array = np.array([len(i) for i in mask_path])
            mask_path = mask_path[np.argmin(mask_array)]
        else:
            mask_path = mask_path[0]
        mask = tf.imread(mask_path).astype('int')
        if not save:
            filename = None

        ## add the mask in a place where it can be recornized by squidpy. The "image" is initialized as a zero-matrix
                # so that the background of the cell is dark.
        dictionary_to_segmentation = {'images':{'segmentation':mask, 'hires':np.zeros([*mask.shape, 3])}}
        with_library_key  = {'filename' : dictionary_to_segmentation}
        subset_anndata.uns['spatial'] = with_library_key
        subset_anndata.obs['cell_id'] = subset_anndata.obs.reset_index().index + 1  # the masks themselves are not 0-indexed, so +1
        subset_anndata.obs['library'] = 'filename'
        subset_anndata.obs['library'] = subset_anndata.obs['library'].astype('category')

        figure = plt.figure()
        ax = plt.gca()
        sq.pl.spatial_segment(subset_anndata, seg_cell_id = 'cell_id', library_key = 'library', color = clustering, seg = True, ax = ax)
        if filename is not None:
            filename = filename + ".png"
            figure.savefig(self.save_cell_maps_dir + "/" + filename, bbox_inches = "tight")
        plt.close()
        return figure, filename
    
    def plot_all_cell_maps(self, clustering: str = "merging"):
        '''
        Plots all the cell maps possible in the dataset
        '''
        for i in self.exp.data.obs['sample_id'].unique():
            self.plot_cell_map(sample_id = i, clustering = clustering, save = True)



# ruff: noqa
'''
This file is an "vendorizing" of the fcsy package: https://github.com/nehcgnay/fcsy

(
MIT License

Copyright (c) 2018, yc

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
)

Changes:
    -- add __all__ (for the sake of autoapi docs -- I don't really want to include any of the vendor files)
    -- Only kept the part needed for PalmettoBUG (DataFrame and its required code)
    -- merged all necessary files into this one
    -- Removing bespoke typing hints (._typing import commented out, all uses of them deleted in main text)
    -- removed now unused / duplicate imports



'''
__all__ = []


__version__ = "0.10.0"

from typing import Union
from io import BytesIO, StringIO
import struct
import re
from abc import abstractmethod, ABC
from contextlib import contextmanager
from functools import update_wrapper, partial

import numpy as np
import pandas as pd

#from ._typing import ReadFcsBuffer, WriteFcsBuffer, UpdateFcsBuffer
#from ._typing import ReadFcsBuffer as ReadFcsBufferType
#from ._typing import WriteFcsBuffer as WriteFcsBufferType

__all__ = ["DataFrame"]

try:
    import boto3  # type: ignore
except ImportError:
    boto3 = None

class ReadFcsBuffer(ABC):
    @abstractmethod
    def seek(self, position: int):
        pass

    @abstractmethod
    def read(self, number: int):
        pass

    @abstractmethod
    def close(self):
        pass


class WriteFcsBuffer(ABC):
    @abstractmethod
    def seek(self, position: int):
        pass

    @abstractmethod
    def write(self, __b: str):
        pass

    @abstractmethod
    def tell(self):
        pass

    @abstractmethod
    def close(self):
        pass


class S3ReadBuffer(ReadFcsBuffer):
    def __init__(self, path, bucket):
        if boto3 is None:
            raise ImportError(
                "This module requires boto3. Please use 'pip install fcsy[boto3] or install boto3 manually"
            )
        super().__init__()
        self.path = path
        self.bucket = bucket
        self.position = 0
        self.s3 = boto3.client("s3")

    def seek(self, position: int):
        self.position = position

    def read(self, number: int):
        resp = self.s3.get_object(
            Bucket=self.bucket,
            Key=self.path,
            Range=f"bytes={self.position}-{self.position+number}",
        )
        self.position += number
        return resp["Body"].read(number)

    def close(self):
        pass


class S3WriteBuffer(WriteFcsBuffer):
    def __init__(self, path, bucket):
        if boto3 is None:
            raise ImportError(
                "This module requires boto3. Please use 'pip install fcsy[boto3] or install boto3 manually"
            )
        super().__init__()
        self.path = path
        self.bucket = bucket
        self.s3 = boto3.client("s3")
        self.buffer = BytesIO()

    def seek(self, position: int):
        self.buffer.seek(position)

    def tell(self):
        return self.buffer.tell()

    def write(self, bytes):
        self.buffer.write(bytes)

    def close(self):
        self.seek(0)
        self.s3.upload_fileobj(self.buffer, self.bucket, self.path)


def read_path(path):
    if path.startswith("s3://"):
        match = re.match(r"s3://(.*?)/(.*?)$", path)
        return {
            "mode": "s3",
            "contents": {"bucket": match.group(1), "key": match.group(2)},
        }
    else:
        return {"mode": "local", "contents": {"path": path}}


def create_open_func(
    parser_class,
    *args,
    **kwargs,
):
    @contextmanager
    def func(path, *func_args, **func_kwargs):
        obj = parser_class(path, *args, **kwargs)
        try:
            yield obj
        finally:
            obj.close()

    return func


class Bufferize:
    def __init__(self, func, mode="rb") -> None:
        update_wrapper(self, func)
        self.func = func
        self.mode = mode

    def __get__(self, obj, objtype):
        return partial(self.__call__, obj)

    def __call__(self, obj, filepath_or_buffer, *args, **kwargs):
        if type(filepath_or_buffer) is str:
            path = read_path(filepath_or_buffer)
            if path["mode"] == "s3":
                if self.mode == "rb":
                    buffer_class = S3ReadBuffer
                elif self.mode == "wb":
                    buffer_class = S3WriteBuffer
                else:
                    raise ValueError("invalid s3 buffer mode")

                open_func = create_open_func(
                    buffer_class, bucket=path["contents"]["bucket"]
                )
                with open_func(path["contents"]["key"], self.mode) as fp:
                    return self.func(obj, fp, *args, **kwargs)
            else:
                with open(filepath_or_buffer, self.mode) as fp:
                    return self.func(obj, fp, *args, **kwargs)
        else:
            return self.func(obj, filepath_or_buffer, *args, **kwargs)


def bufferize(func=None, mode="rb"):
    if func:
        return Bufferize(func)
    else:

        def wrapper(func):
            return Bufferize(func, mode=mode)

        return wrapper
    
class HeaderSegment:
    ranges = {
        "version": (0, 5),
        "text_start": (10, 17),
        "text_end": (18, 25),
        "data_start": (26, 33),
        "data_end": (34, 41),
        "analysis_start": (42, 49),
        "analysis_end": (50, 57),
    }

    def __init__(
        self,
        text_start: int,
        text_end,
        data_start,
        data_end,
        analysis_start=0,
        analysis_end=0,
        ver: str = "FCS3.1",
    ) -> None:
        self.ver = ver
        self.text_start = text_start
        self.text_end = text_end
        self.data_start = data_start
        self.data_end = data_end
        self.analysis_start = analysis_start
        self.analysis_end = analysis_end

    def asdict(self):
        return {
            "text_start": self.text_start,
            "text_end": self.text_end,
            "data_start": self.data_start,
            "data_end": self.data_end,
            "analysis_start": self.analysis_start,
            "analysis_end": self.analysis_end,
        }

    @classmethod
    def from_string(cls, s: bytes):
        s = s.decode("UTF-8")
        v = dict()
        for key, value in cls.ranges.items():
            v[key] = s[value[0] : value[1] + 1].strip()
            if key != "version":
                v[key] = int(v[key])

        ver = v.pop("version")
        return HeaderSegment(
            v["text_start"],
            v["text_end"],
            v["data_start"],
            v["data_end"],
            v["analysis_start"],
            v["analysis_end"],
            ver=ver,
        )

    def _write_locations(self, sp: StringIO, type: str):
        sp.seek(self.ranges[f"{type}_start"][0])
        start_str = str(self.asdict()[f"{type}_start"])
        end_str = str(self.asdict()[f"{type}_end"])

        if len(start_str) > 8 or len(end_str) > 8:
            start_str = "0"
            end_str = "0"

        sp.write((8 - len(start_str)) * " " + start_str)
        sp.seek(self.ranges[f"{type}_end"][0])
        sp.write((8 - len(end_str)) * " " + end_str)

    def to_string(self):
        header = None
        with StringIO() as sp:
            sp.write(self.ver)
            sp.write(" " * 4)
            for t in ["text", "data", "analysis"]:
                self._write_locations(sp, t)
            header = sp.getvalue()

        return header


class TextSegment:
    def __init__(
        self,
        tot: int,
        pnn: list,
        pns: list,
        pnr: list,
        pne: list = None,
        mode: str = "L",
        next_data: int = 0,
        byte_order: str = "1,2,3,4",
        text_start=256,
        delim="/",
        datatype="F",
        data_start=None,
        data_end=None,
        analysis_start=None,
        analysis_end=None,
        stext_start=None,
        stext_end=None,
    ) -> None:
        self.tot = tot
        self.pnn = pnn
        self.pns = pns
        self.pnb = "32"
        self.pne = pne
        self.pnr = pnr
        self.mode = mode
        self.next_data = next_data
        self.byte_order = "1,2,3,4"
        self.text_start = text_start
        self.delim = delim
        self.datatype = datatype
        self.endian = {"1,2,3,4": "<", "4,3,2,1": ">"}[byte_order]
        self._data_start = data_start
        self._data_end = data_end
        self._analysis_start = analysis_start
        self._analysis_end = analysis_end
        self._stext_start = stext_start
        self._stext_end = stext_end
        self.text = self.build()

    def build(self):
        text = {}
        text["BYTEORD"] = self.byte_order
        text["DATATYPE"] = self.datatype
        text["MODE"] = self.mode
        text["NEXTDATA"] = str(self.next_data)
        text["BEGINANALYSIS"] = (
            str(self._analysis_start) if self._analysis_start is not None else "0"
        )
        text["ENDANALYSIS"] = (
            str(self._analysis_end) if self._analysis_end is not None else "0"
        )
        text["BEGINDATA"] = (
            str(self._data_start) if self._data_start is not None else "0"
        )
        text["ENDDATA"] = str(self._data_end) if self._data_end is not None else "0"
        text["BEGINSTEXT"] = (
            str(self._stext_start) if self._stext_start is not None else "0"
        )
        text["ENDSTEXT"] = str(self._stext_end) if self._stext_end is not None else "0"

        text["PAR"] = str(len(self.pnn))
        text["TOT"] = str(self.tot)

        for i, _ in enumerate(self.pnn):
            text["P%dB" % (i + 1)] = self.pnb
            text["P%dE" % (i + 1)] = self.pne
            text["P%dR" % (i + 1)] = str(int(self.pnr[i]))
            text["P%dN" % (i + 1)] = self.pnn[i]
            text["P%dS" % (i + 1)] = self.pns[i]

        if self._data_start is None and self._data_end is None:
            max_text_end = self.cal_max_text_ends(text, self.delim, self.text_start)
            datasize = 4 * len(self.pnn) * self.tot  # 4 is bytes to hold float
            data_start, data_end = self.cal_datapos(max_text_end + 1, datasize)
            text["BEGINDATA"] = data_start
            text["ENDDATA"] = data_end
        return text

    @property
    def data_start(self):
        return int(self.text["BEGINDATA"])

    @property
    def data_end(self):
        return int(self.text["ENDDATA"])

    @property
    def text_end(self):
        return len(self.dict_to_string(self.text, self.delim)) + self.text_start - 1

    @classmethod
    def cal_max_text_ends(cls, text, delim, text_start):
        s = cls.dict_to_string(
            {
                **text,
                "BEGINANALYSIS": "99999999",
                "ENDANALYSIS": "99999999",
                "BEGINDATA": "99999999",
                "ENDDATA": "99999999",
                "BEGINSTEXT": "99999999",
                "ENDSTEXT": "99999999",
            },
            delim,
        )
        return text_start + len(s) - 1

    def to_string(self):
        return self.dict_to_string(self.text, self.delim)

    def update_pns(self, mappings):
        for k, v in mappings.items():
            self.pns[self.pns.index(k)] = v
        self.text = self.build()

    def update_pnn(self, mappings):
        for k, v in mappings.items():
            self.pnn[self.pnn.index(k)] = v
        self.text = self.build()

    @classmethod
    def from_string(self, s):
        delim = s[:1]
        keyvalarray = s[1:].split(delim)
        for num, i in enumerate(keyvalarray):
            try:
                keyvalarray[num] = i.decode("UTF-8")
            except UnicodeDecodeError:
                # print(num, i)
                pass

        vars = {}
        for k, v in zip(keyvalarray[::2], keyvalarray[1::2]):
            vars[k[1:]] = v

        long_channels = list()
        short_channels = list()
        max_values = list()
        for n in range(int(vars.pop("PAR"))):
            short_channels.append(vars.pop("P%dN" % (n + 1)))
            long_channels.append(
                vars.pop("P%dS" % (n + 1)) if "P%dS" % (n + 1) in vars else " "
            )
            max_values.append(int(vars.pop("P%dR" % (n + 1))))

        return TextSegment(
            int(vars["TOT"]),
            short_channels,
            long_channels,
            max_values,
            mode=vars["MODE"],
            next_data=vars["NEXTDATA"],
            byte_order=vars["BYTEORD"],
            delim=delim.decode("UTF-8"),
            datatype=vars["DATATYPE"],
            data_start=vars["BEGINDATA"],
            data_end=vars["ENDDATA"],
            analysis_start=vars["BEGINANALYSIS"],
            analysis_end=vars["ENDANALYSIS"],
            stext_start=vars["BEGINSTEXT"],
            stext_end=vars["ENDSTEXT"],
        )

    @classmethod
    def dict_to_string(cls, dict_, delim):
        # print('convert', dict_)
        s = delim
        for i in dict_:
            # print(type(i), type(dict_[i]), type(delim), i, dict_[i])
            s += "$%s%s%s%s" % (i, delim, dict_[i], delim)
        return s

    @classmethod
    def cal_datapos(cls, textend, datasize):
        start = 0
        end = 0
        start_new = textend + 1
        end_new = start_new + datasize - 1

        while True:
            inc_start = len(str(start_new)) - len(str(start))
            inc_end = len(str(end_new)) - len(str(end))
            start = start_new
            end = end_new

            if inc_start + inc_end > 0:
                start_new = inc_start + inc_end + start
                end_new = start_new + datasize - 1
            else:
                break

        return start, end


class DataSegment:
    def __init__(
        self, values: np.ndarray, datatype, num_dims, num_rows, endian
    ) -> None:
        self.values = values
        self.datatype = datatype
        self.num_dims = num_dims
        self.num_rows = num_rows
        self.endian = endian

    def to_string(self):
        format_ = self.endian + self.datatype.lower()
        return self.values.astype(format_).tobytes()

    @classmethod
    def from_string(cls, s: str, datatype, num_dims, num_rows, endian):
        data = BytesIO(s)
        format_ = endian + str(num_dims) + datatype.lower()
        datasize = struct.calcsize(format_)
        events = []
        for e in range(num_rows):
            read_data = data.read(datasize)
            event = struct.unpack(format_, read_data)
            events.append(event)
        values = np.array(events, dtype=np.float32 if datatype == "F" else np.float64)
        return DataSegment(values, datatype, num_dims, num_rows, endian)


class Fcs:
    def __init__(
        self,
        values: np.ndarray,
        short_channels: list,
        long_channels: list = None,
        ver: str = "FCS3.1",
        mode: str = "L",
        byte_order: str = "1,2,3,4",
        text_start: int = 256,
        datatype: str = "F",
    ) -> None:
        self.values = values
        self.short_channels = short_channels
        if long_channels is not None:
            self.long_channels = long_channels
        else:
            self.long_channels = [" " for _ in range(len(short_channels))]
        self.ver = ver
        self.mode = mode
        self.byte_order = byte_order

        self.tseg = TextSegment(
            values.shape[0],
            self.short_channels,
            self.long_channels,
            values.max(axis=0),
            mode=mode,
            byte_order=byte_order,
            text_start=text_start,
            datatype=datatype,
        )

        self.hseg = HeaderSegment(
            self.tseg.text_start,
            self.tseg.text_end,
            self.tseg.data_start,
            self.tseg.data_end,
            ver=self.ver,
        )

        self.dseg = DataSegment(
            values, datatype, values.shape[1], values.shape[0], self.tseg.endian
        )

    @classmethod
    @bufferize
    def from_file(cls, filepath_or_buffer):
        header = HeaderSegment.from_string(filepath_or_buffer.read(58))
        filepath_or_buffer.seek(header.text_start)
        fseg = TextSegment.from_string(
            filepath_or_buffer.read(header.text_end - header.text_start + 1)
        )
        data_start = header.data_start if header.data_start != 0 else fseg.data_start
        data_end = header.data_end if header.data_end != 0 else fseg.data_end
        filepath_or_buffer.seek(data_start)
        data = DataSegment.from_string(
            filepath_or_buffer.read(data_end - data_start + 1),
            fseg.datatype,
            len(fseg.pnn),
            fseg.tot,
            fseg.endian,
        )

        return Fcs(data.values, fseg.pnn, fseg.pns)

    @classmethod
    @bufferize
    def read_data_segment(cls, filepath_or_buffer, hseg, tseg):
        data_start = hseg.data_start if hseg.data_start != 0 else tseg.data_start
        data_end = hseg.data_end if hseg.data_end != 0 else tseg.data_end
        filepath_or_buffer.seek(data_start)
        return DataSegment.from_string(
            filepath_or_buffer.read(data_end - data_start + 1),
            tseg.datatype,
            len(tseg.pnn),
            tseg.tot,
            tseg.endian,
        )

    @classmethod
    @bufferize
    def read_header_segment(cls, filepath_or_buffer):
        filepath_or_buffer.seek(0)
        return HeaderSegment.from_string(filepath_or_buffer.read(58))

    @classmethod
    @bufferize
    def read_text_segment(cls, filepath_or_buffer):
        header = HeaderSegment.from_string(filepath_or_buffer.read(58))
        filepath_or_buffer.seek(header.text_start)
        fseg = TextSegment.from_string(
            filepath_or_buffer.read(header.text_end - header.text_start + 1)
        )

        return fseg

    @classmethod
    def write_bytes(cls, f, s):
        if not type(s) == bytes:
            f.write(s.encode("UTF-8"))
        else:
            f.write(s)

    @classmethod
    @bufferize(mode="rb+")
    def write_text_segment(cls, filepath_or_buffer, text_segment):
        filepath_or_buffer.seek(58)
        cls.write_bytes(
            filepath_or_buffer,
            " " * (text_segment.text_start - filepath_or_buffer.tell()),
        )
        cls.write_bytes(filepath_or_buffer, text_segment.to_string())
        cls.write_bytes(
            filepath_or_buffer,
            " " * (text_segment.data_start - filepath_or_buffer.tell()),
        )

    @classmethod
    @bufferize(mode="rb+")
    def write_header_segment(cls, filepath_or_buffer, header_segment):
        filepath_or_buffer.seek(0)
        cls.write_bytes(filepath_or_buffer, header_segment.to_string())

    @bufferize(mode="wb")
    def export(self, filepath_or_buffer):
        self.write_bytes(filepath_or_buffer, self.hseg.to_string())
        self.write_text_segment(filepath_or_buffer, self.tseg)
        self.write_bytes(filepath_or_buffer, self.dseg.to_string())

    @property
    def count(self):
        return self.tseg.tot

class DataFrame(pd.DataFrame):
    @property
    def _constructor(self):
        return DataFrame

    def to_fcs(self, filepath_or_buffer: Union[str]):

        """
        write fcs.

        :param filepath_or_buffer:
            String, or file-like object implementing a ``write()`` function.
            The string could be a s3 url with the format:
            ``s3://{bucket}/{key}``.
        :type param filepath_or_buffer: str or file-like object
        """

        long_channels = None
        if isinstance(self.columns, pd.MultiIndex):
            short_channels = self.columns.get_level_values("short").values
            long_channels = self.columns.get_level_values("long").values
        else:
            short_channels = self.columns

        Fcs(self.values, short_channels, long_channels).export(filepath_or_buffer)

    @classmethod
    def from_fcs(
        cls, filepath_or_buffer: Union[str], channel_type: str = "short"
    ):
        """
        Read dataframe from fcs.

        :param filepath_or_buffer: str or file-like object.
            String, or file-like object implementing a ``read()`` function.
            The string could be a s3 url with the format:
            ``s3://{bucket}/{key}``.
        :type filepath_or_buffer: str or file-like object
        :param channel_type: {"short", "long", "multi"}, defaults to "short".
            "short" and "long" refer to short ($PnN) and long ($PnS) name of parameter n, respectively.
            See FCS3.1 data standard for detailed explanation.
        :type channel_type: str, optional
        :return: the dataframe contains the fcs channels and data
        :rtype: DataFrame
        """

        fcs = Fcs.from_file(filepath_or_buffer)
        colmap = {
            "short": fcs.short_channels,
            "long": fcs.long_channels,
            "multi": pd.MultiIndex.from_tuples(
                list(zip(fcs.short_channels, fcs.long_channels)),
                names=["short", "long"],
            ),
        }
        return DataFrame(fcs.values, columns=colmap[channel_type])